#  Copyright 2020 Regents of the University of Minnesota.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Events service client API and wrapper classes.

"""

import collections
import threading
import uuid
from typing import Iterator, List, Dict, MutableMapping, Generic, TypeVar, NamedTuple, \
    ContextManager, Iterable, Optional, Sequence, Union, Mapping, TYPE_CHECKING, Callable

import grpc
from grpc_health.v1 import health_pb2_grpc, health_pb2

from mtap import _config, _discovery, constants
from mtap.api.v1 import events_pb2, events_pb2_grpc
from mtap.data import _base, _labels, _label_adapters

if TYPE_CHECKING:
    import mtap
    import mtap.data as data

L = TypeVar('L', bound='data.Label')


class Event:
    """An object for interacting with a specific event locally or on the events service.

    The Event object functions as a map from string document names to :obj:`Document` objects that
    can be used to access document data from the events server.

    Keyword Args:
        event_id (~typing.Optional[str]):
            A globally-unique identifier for the event, or omit / none for a random UUID.
        client (~typing.Optional[EventsClient]):
            A client for an events service to push any changes to the event to.
        only_create_new (bool): Fails if the event already exists on the events service.

    Examples:
        >>> with EventsClient() as c, Event(event_id='id', client=c) as event:
        >>>     # use event
        >>>     ...
    """
    __slots__ = ('_event_id', '_client', '_lock', 'default_adapters', '_documents', '_metadata',
                 '_binaries', '_event_service_instance_id')

    def __init__(self, *, event_id: Optional[str] = None,
                 event_service_instance_id: Optional[str] = None,
                 client: Optional['mtap.EventsClient'] = None,
                 only_create_new: bool = False,
                 default_adapters: Optional[Mapping[str, 'data.ProtoLabelAdapter']] = None):
        self._event_id = event_id or str(uuid.uuid4())
        self._event_service_instance_id = event_service_instance_id
        self._lock = threading.RLock()
        self.default_adapters = default_adapters or {}
        self._client = client
        if self._client is not None:
            self._client = client.get_local_instance(self._event_service_instance_id)
            self._event_service_instance_id = self._client.instance_id
            self._client.open_event(self._event_id, only_create_new=only_create_new)

    @property
    def client(self) -> Optional['mtap.EventsClient']:
        return self._client

    @property
    def event_id(self) -> str:
        """str: The globally unique identifier for this event."""
        return self._event_id

    @property
    def event_service_instance_id(self) -> str:
        """str: The unique instance identifier for this event's paired event service."""
        return self._event_service_instance_id

    @property
    def documents(self) -> MutableMapping[str, 'mtap.Document']:
        """~typing.MutableMapping[str, Document]: A mutable mapping of strings to :obj:`Document`
        objects that can be used to query and add documents to the event."""
        try:
            return self._documents
        except AttributeError:
            self._documents = _Documents(self, self.client)
            return self._documents

    @property
    def metadata(self) -> MutableMapping[str, str]:
        """~typing.MutableMapping[str, str]: A mutable mapping of strings to strings that can be
        used to query and add metadata to the event."""
        try:
            return self._metadata
        except AttributeError:
            self._metadata = _Metadata(self, self.client)
            return self._metadata

    @property
    def binaries(self) -> MutableMapping[str, bytes]:
        """~typing.MutableMapping[str, str]: A mutable mapping of strings to bytes that can be used
        to query and add binary data to the event."""
        try:
            return self._binaries
        except AttributeError:
            self._binaries = _Binaries(self, self.client)
            return self._binaries

    @property
    def created_indices(self) -> Dict[str, List[str]]:
        """~typing.Dict[str, ~typing.List[str]]: A mapping of document names to a list of the names
        of all the label indices that have been added to that document"""
        return {document_name: document.created_indices
                for document_name, document in self.documents.items()}

    def close(self):
        """Closes this event. Lets the event service know that we are done with the event,
        allowing to clean up the event if no other clients have open leases to it."""
        if self.client is not None:
            self.release_lease()

    def create_document(self, document_name: str, text: str) -> 'mtap.Document':
        """Adds a document to the event keyed by `document_name` and
        containing the specified `text`.

        Args:
            document_name (str): The event-unique identifier for the document, example: 'plaintext'.
            text (str):
                The content of the document. This is a required field, document text is final and
                immutable, as changing the text would very likely invalidate any labels on the
                document.

        Returns:
            Document: The added document.

        Examples:

            >>> event = Event()
            >>> document = event.create_document('plaintext', text="The text of the document.")

        """
        if not isinstance(text, str):
            raise ValueError('text is not string.')
        document = Document(document_name, text=text, event=self)
        self.documents[document_name] = document
        return document

    def add_document(self, document: 'mtap.Document'):
        """Adds the document to this event, first uploading to events service if this event has a
        client connection to the events service.

        Args:
            document (~mtap.Document): The document to add to this event.

        Examples:

            >>> event = Event()
            >>> document = Document('plaintext', text="The text of the document.")
            >>> event.add_document(document)

        """
        if document._event is not None:
            raise ValueError('The document already exists on an event')
        document._event = self
        document._client = self.client
        document._event_id = self.event_id
        self.documents[document.document_name] = document

    def __enter__(self) -> 'Event':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def add_created_indices(self, created_indices):
        for k, v in created_indices.items():
            try:
                doc = self._documents[k]
                doc.add_created_indices(v)
            except KeyError:
                pass

    def lease(self):
        if self.client is not None:
            self.client.open_event(self.event_id, only_create_new=False)

    def release_lease(self):
        if self.client is not None:
            self.client.close_event(self.event_id)


class _WaitingIndex(NamedTuple('_WaitingIndex',
                               [('index_name', str),
                                ('labels', List['mtap.GenericLabel']),
                                ('waiting_on', List[int])])):
    """A label index that is waiting on some labels to be staticized before uploading.
    """


class Document:
    """An object for interacting with text and labels stored on an :class:`Event`.

    Documents are keyed by their name, and pipelines can store different pieces of
    related text on a single processing event using multiple documents. An example would be storing
    the text of one language on one document, and a translation on another, or storing the
    rtf or html encoding on one document, and the parsed plaintext on another document.

    Both the document text and any added label indices are immutable. This is to enable
    parallelization and distribution of processing, and to prevent changes to the dependency graph
    of label indices and text, which can make debugging difficult.

    Args:
        document_name (str): The document name identifier.

    Keyword Args:
        text (~typing.Optional[str]):
            The document text, can be omitted if this is an existing document and text needs to be
            retrieved from the events service.
        event (~typing.Optional[Event]):
            The parent event of this document. If the event has a client, then that client will be
            used to share changes to this document with all other clients of the Events service. In
            that case, text should only be specified if it is the known existing text of the
            document.

    Examples:
        Local document:

        >>> document = Document('plaintext', text='Some document text.')

        Existing distributed object:

        >>> with EventsClient(address='localhost:8080') as client, \\
        >>>      Event(event_id='1', client=client) as event:
        >>>     document = event.documents['plaintext']
        >>>     document.text
        'Some document text fetched from the server.'

        New distributed object:

        >>> with EventsClient(address='localhost:8080') as client, \\
        >>>      Event(event_id='1', client=client) as event:
        >>>     document = Document('plaintext', text='Some document text.')
        >>>     event.add_document(document)

        or

        >>> with EventsClient(address='localhost:8080') as client, \\
        >>>      Event(event_id='1', client=client) as event:
        >>>     document = event.create_document('plaintext', text='Some document text.')

    """
    __slots__ = ('_document_name', '_event', '_client', '_event_id', '_text', '_labelers',
                 '_created_indices', '_waiting_indices', 'default_adapters', '_labels')

    def __init__(self,
                 document_name: str,
                 *,
                 text: Optional[str] = None,
                 event: Optional[Event] = None,
                 default_adapters: Optional[Mapping[str, 'data.ProtoLabelAdapter']] = None):
        if not isinstance(document_name, str):
            raise TypeError('Document name is not string.')
        self._document_name = document_name
        self._event = event
        self._client = None
        self._event_id = None
        if event is not None:
            self._client = event.client
            self._event_id = event.event_id
        self._text = text
        self._labelers = []
        self._created_indices = []
        self._waiting_indices = []
        self.default_adapters = default_adapters or {}
        if self._client is None and text is None:
            raise ValueError('Document without text or an event with a client to fetch the text '
                             'from.')

    @property
    def event(self) -> Event:
        """Event: The parent event of this document."""
        return self._event

    @property
    def document_name(self) -> str:
        """str: The unique identifier for this document on the event."""
        return self._document_name

    @property
    def text(self):
        """str: The document text."""
        if self._text is None and self._client is not None:
            self._text = self._client.get_document_text(self._event_id, self._document_name)
        return self._text

    @property
    def created_indices(self) -> List[str]:
        """~typing.List[str]: A list of all of the label index names that have created on this
                document using a labeler either locally or by remote pipeline components invoked on
                this document."""
        return list(self._created_indices)

    @property
    def labels(self) -> Mapping[str, 'data.LabelIndex']:
        try:
            return self._labels
        except AttributeError:
            self._labels = _LabelIndices(self)
            return self._labels

    def get_label_index(self,
                        label_index_name: str) -> 'data.LabelIndex[Union[mtap.GenericLabel, L]]':
        """Gets the document's label index with the specified key.

        Will fetch from the events service if it is not cached locally if the document has an event
        with a client. Uses the `label_adapter` argument to perform unmarshalling from the proto
        message if specified.

        Args:
            label_index_name (str): The name of the label index to get.

        Returns:
            LabelIndex: The requested label index.
        """
        return self.labels[label_index_name]

    def get_labeler(self,
                    label_index_name: str,
                    *, distinct: Optional[bool] = None):
        """Creates a function that can be used to add labels to a label index.

        Args:
            label_index_name (str): A document-unique identifier for the label index to be created.
            distinct (~typing.Optional[bool]):
                Optional, if using generic labels, whether to use distinct generic labels or
                non-distinct generic labels, will default to False.

        Returns:
            Labeler: A callable when used in conjunction with the 'with' keyword will automatically
            handle uploading any added labels to the server.

        Examples:
            >>> with document.get_labeler('sentences', distinct=True) as labeler:
            >>>     labeler(0, 25, sentence_type='STANDARD')
            >>>     sentence = labeler(26, 34)
            >>>     sentence.sentence_type = 'FRAGMENT'
        """
        if label_index_name in self._labelers:
            raise KeyError("Labeler already in use: " + label_index_name)
        label_adapter = self.get_default_adapter(label_index_name, distinct)

        labeler = Labeler(self._client, self, label_index_name, label_adapter)
        self._labelers.append(label_index_name)
        return labeler

    def add_labels(self,
                   label_index_name: str,
                   labels: Sequence[Union['data.Label', L]],
                   *, distinct: Optional[bool] = None,
                   label_adapter: Optional['data.ProtoLabelAdapter'] = None):
        """Skips using a labeler and adds the sequence of labels as a new label index.

        Args:
            label_index_name (str): The name of the label index.
            labels (~typing.Sequence[Label]): The labels to add.
            distinct (~typing.Optional[bool]):
                Whether the index is distinct or non-distinct.
            label_adapter (mtap.label_adapters.ProtoLabelAdapter): A label adapter to use.

        Returns:
            LabelIndex: The new label index created from the labels.
        """
        if label_index_name in self.labels:
            raise KeyError("Label index already exists with name: " + label_index_name)
        if label_adapter is None:
            if distinct is None:
                distinct = False
            label_adapter = self.get_default_adapter(label_index_name, distinct)
        labels, waiting_on = _labels._staticize(labels, self, label_index_name)
        if len(self._waiting_indices) > 0:
            label_ids = {id(label) for label in labels}
            self._check_waiting_indices(label_ids)

        if len(waiting_on) > 0:
            self._waiting_indices.append((label_index_name, labels, label_adapter, waiting_on))
        else:
            self._finalize_labels(label_adapter, label_index_name, labels)

    def _check_waiting_indices(self, updated_ids):
        waiting_indices = []
        for label_index_name, labels, label_adapter, waiting_on in self._waiting_indices:
            still_waiting = waiting_on.difference(updated_ids)
            if len(still_waiting) == 0:
                self._finalize_labels(label_adapter, label_index_name, labels)
            else:
                waiting_indices.append((label_index_name, labels, label_adapter, still_waiting))
        self._waiting_indices = waiting_indices

    def _finalize_labels(self, label_adapter, label_index_name, labels):
        label_adapter.store_references(labels)
        if self._client is not None:
            self._client.add_labels(event_id=self.event.event_id,
                                    document_name=self.document_name,
                                    index_name=label_index_name,
                                    labels=labels,
                                    adapter=label_adapter)
        self._created_indices.append(label_index_name)
        index = label_adapter.create_index(labels)
        self.labels.add_to_cache(label_index_name, index)
        return index

    def get_default_adapter(self, label_index_name: str,
                            distinct: Optional[bool] = None) -> Optional['data.ProtoLabelAdapter']:
        try:
            return self.default_adapters[label_index_name]
        except KeyError:
            pass
        try:
            return self.event.default_adapters[label_index_name]
        except (AttributeError, KeyError):
            return _label_adapters.DISTINCT_GENERIC_ADAPTER if distinct \
                else _label_adapters.GENERIC_ADAPTER

    def add_created_indices(self, created_indices: Iterable[str]):
        # Internal, used by the pipeline to add any indices created remotely to the
        # "created_indices" on a local document.
        return self._created_indices.extend(created_indices)


class Labeler(Generic[L], ContextManager['Labeler']):
    """Object provided by :func:`~'mtap.Document'.get_labeler` which is responsible for adding labels
    to a label index on a document.
    """
    __slots__ = ('_client', '_document', '_label_index_name', '_label_adapter', 'is_done',
                 '_current_labels', '_lock')

    def __init__(self,
                 client: 'mtap.EventsClient',
                 document: Document,
                 label_index_name: str,
                 label_adapter: 'data.ProtoLabelAdapter[L]'):
        self._client = client
        self._document = document
        self._label_index_name = label_index_name
        self._label_adapter = label_adapter
        self.is_done = False
        self._current_labels = []
        self._lock = threading.Lock()

    def __call__(self, *args, **kwargs) -> L:
        """Calls the constructor for the label type adding it to the list of labels to be uploaded.

        Args:
            args: Arguments passed to the label type's constructor.
            kwargs: Keyword arguments passed to the label type's constructor.

        Returns:
            Label: The object that was created by the label type's constructor.

        Examples:
            >>> labeler(0, 25, some_field='some_value', x=3)
            GenericLabel(start_index=0, end_index=25, some_field='some_value', x=3)
        """
        label = self._label_adapter.create_label(*args, document=self._document, **kwargs)
        self._current_labels.append(label)
        return label

    def __enter__(self) -> 'data.Labeler':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            return False
        self.done()

    def done(self):
        """Finalizes the label index, uploads the added labels to the events service.

        Normally called automatically on exit from a context manager block, but can be manually
        invoked if the labeler is not used in a context manager block."""
        with self._lock:
            if self.is_done:
                return
            self.is_done = True
            self._document.add_labels(self._label_index_name, self._current_labels,
                                      label_adapter=self._label_adapter)


def create_channel(address):
    config = _config.Config()
    options = config.get('grpc.events_channel_options', {})
    channel = grpc.insecure_channel(address, options=list(options.items()))

    health = health_pb2_grpc.HealthStub(channel)
    hcr = health.Check(health_pb2.HealthCheckRequest(service=constants.EVENTS_SERVICE_NAME))
    if hcr.status != health_pb2.HealthCheckResponse.SERVING:
        raise ValueError('Failed to connect to events service. Status:')
    return channel


class _ClientPool:
    __slots__ = ('_ptr', 'clients', 'instance_id_to_delegate')

    def __init__(self, channel_factory, addresses):
        self._ptr = 0
        self.clients = [EventsClient(address=addresses, channel_factory=channel_factory,
                                     _pool=self, _channel=channel_factory(a))
                        for a in addresses]
        self.instance_id_to_delegate = {d.instance_id: d for d in self.clients}

    def get_instance(self, instance_id):
        try:
            return self.instance_id_to_delegate[instance_id]
        except KeyError:
            i = self.clients[self._ptr]
            self._ptr = (self._ptr + 1) % len(self.clients)
            return i


class EventsClient:
    """A client object for interacting with the events service.

    Normally, users shouldn't have to use any of the methods on this object, as they are invoked by
    the globally distributed object classes of :obj:`Event`, :obj:`Document`, and :obj:`~data.Labeler`.

    Keyword Args:
        address (~typing.Optional[str]): The events service target e.g. 'localhost:9090' or
            omit/None to use service discovery.
        stub (~typing.Optional[~mtap.api.v1.events_pb2_grpc.EventsStub]): An existing events service
            client gRPC stub to use.

    Examples:
        >>> with EventsClient(address='localhost:50000') as client, \\
        >>>      Event(event_id='1', client=client) as event:
        >>>     document = event.create_document(document_name='plaintext',
        >>>                                      text='The quick brown fox jumps over the lazy dog.')
    """
    __slots__ = ('addresses', 'channel_factory', '_pool', '_channel', 'stub', '_instance_id',
                 '_is_open')

    def __init__(self, address: Optional[Union[str, Iterable[str]]],
                 channel_factory: Callable[[str], grpc.Channel] = None,
                 _pool: Optional[_ClientPool] = None,
                 _channel: Optional[grpc.Channel] = None):
        if address is None:
            discovery = _discovery.Discovery(_config.Config())
            address = discovery.discover_events_service('v1')
        if isinstance(address, str):
            self.addresses = [address]
        elif isinstance(address, Iterable):
            self.addresses = list(address)
        else:
            raise ValueError("Unrecognized address type: " + str(type(address)))
        self.channel_factory = channel_factory
        if channel_factory is None:
            self.channel_factory = create_channel
        self._pool = _pool
        if self._pool is None:
            self._pool = _ClientPool(self.channel_factory, self.addresses)
        self._channel = _channel
        if self._channel is not None:
            self.stub = events_pb2_grpc.EventsStub(self._channel)
            r = self.stub.GetEventsInstanceId(events_pb2.GetEventsInstanceIdRequest())
            self._instance_id = r.instance_id
            self._is_open = True

    @property
    def instance_id(self) -> str:
        return self._instance_id

    def __enter__(self) -> 'EventsClient':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __reduce__(self):
        return _create_events_client, (self.addresses, self.channel_factory,)

    def get_local_instance(self, instance_id: Optional[str]) -> 'EventsClient':
        return self._pool.get_instance(instance_id)

    def ensure_open(self):
        if not self._is_open:
            raise ValueError("Client to events service is not open")

    def open_event(self, event_id: str, only_create_new: bool):
        request = events_pb2.OpenEventRequest(event_id=event_id, only_create_new=only_create_new)
        try:
            response = self.stub.OpenEvent(request)
            assert response is not None
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            if e.code() == grpc.StatusCode.ALREADY_EXISTS:
                raise ValueError("Event already exists") from e
            raise e

    def close_event(self, event_id):
        request = events_pb2.CloseEventRequest(event_id=event_id)
        try:
            response = self.stub.CloseEvent(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return response is not None

    def get_all_metadata(self, event_id):
        request = events_pb2.GetAllMetadataRequest(event_id=event_id)
        try:
            response = self.stub.GetAllMetadata(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return response.metadata

    def add_metadata(self, event_id, key, value):
        request = events_pb2.AddMetadataRequest(event_id=event_id, key=key, value=value)
        try:
            response = self.stub.AddMetadata(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return response is not None

    def get_all_binary_data_names(self, event_id: str) -> List[str]:
        request = events_pb2.GetAllBinaryDataNamesRequest(event_id=event_id)
        try:
            response = self.stub.GetAllBinaryDataNames(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return list(response.binary_data_names)

    def add_binary_data(self, event_id: str, binary_data_name: str, binary_data: bytes):
        request = events_pb2.AddBinaryDataRequest(event_id=event_id,
                                                  binary_data_name=binary_data_name,
                                                  binary_data=binary_data)
        try:
            response = self.stub.AddBinaryData(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return response is not None

    def get_binary_data(self, event_id: str, binary_data_name: str) -> bytes:
        request = events_pb2.GetBinaryDataRequest(event_id=event_id,
                                                  binary_data_name=binary_data_name)
        try:
            response = self.stub.GetBinaryData(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return response.binary_data

    def get_all_document_names(self, event_id):
        request = events_pb2.GetAllDocumentNamesRequest(event_id=event_id)
        try:
            response = self.stub.GetAllDocumentNames(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return list(response.document_names)

    def add_document(self, event_id, document_name, text):
        request = events_pb2.AddDocumentRequest(event_id=event_id,
                                                document_name=document_name,
                                                text=text)
        try:
            response = self.stub.AddDocument(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return response is not None

    def get_document_text(self, event_id, document_name):
        request = events_pb2.GetDocumentTextRequest(event_id=event_id,
                                                    document_name=document_name)
        try:
            response = self.stub.GetDocumentText(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return response.text

    def get_label_index_info(self,
                             event_id: str,
                             document_name: str) -> List['data.LabelIndexInfo']:
        request = events_pb2.GetLabelIndicesInfoRequest(event_id=event_id,
                                                        document_name=document_name)
        try:
            response = self.stub.GetLabelIndicesInfo(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        result = []
        for index in response.label_index_infos:
            if index.type == events_pb2.GetLabelIndicesInfoResponse.LabelIndexInfo.GENERIC:
                index_type = _base.LabelIndexType.GENERIC
            elif index.type == events_pb2.GetLabelIndicesInfoResponse.LabelIndexInfo.CUSTOM:
                index_type = _base.LabelIndexType.CUSTOM
            else:
                index_type = _base.LabelIndexType.UNKNOWN
            result.append(_base.LabelIndexInfo(index.index_name, index_type))
        return result

    def add_labels(self, event_id, document_name, index_name, labels, adapter):
        request = events_pb2.AddLabelsRequest(event_id=event_id, document_name=document_name,
                                              index_name=index_name,
                                              no_key_validation=True)
        adapter.add_to_message(labels, request)
        try:
            response = self.stub.AddLabels(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return response is not None

    def get_labels(self, event_id, document_name, index_name, adapter):
        request = events_pb2.GetLabelsRequest(event_id=event_id,
                                              document_name=document_name,
                                              index_name=index_name)
        try:
            response = self.stub.GetLabels(request)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE:
                raise ConnectionError("Failed to connect to events service on addresses: {}".format(self.addresses))
            raise e
        return adapter.create_index_from_response(response)

    def close(self):
        self._is_open = False
        try:
            self._channel.close()
        except AttributeError:
            pass


def _create_events_client(addresses, channel_factory):
    return EventsClient(address=addresses, channel_factory=channel_factory)


class _Documents(MutableMapping[str, Document]):
    __slots__ = ('event', 'event_id', 'client', 'documents')

    def __init__(self, event: Event, client: Optional[EventsClient]):
        self.event = event
        self.event_id = event.event_id
        self.client = client
        self.documents = {}

    def __contains__(self, document_name: str) -> bool:
        if not isinstance(document_name, str):
            return False
        if document_name in self.documents:
            return True
        self._refresh_documents()
        return document_name in self.documents

    def __getitem__(self, document_name) -> 'Document':
        if not isinstance(document_name, str):
            raise KeyError
        try:
            return self.documents[document_name]
        except KeyError:
            pass
        self._refresh_documents()
        return self.documents[document_name]

    def __len__(self) -> int:
        self._refresh_documents()
        return len(self.documents)

    def __iter__(self) -> Iterator[str]:
        self._refresh_documents()
        return iter(self.documents)

    def __setitem__(self, k: str, v: Document) -> None:
        if self.client is not None:
            if not self.client.add_document(self.event_id, k, v.text):
                raise ValueError()
        v._event = self.event
        v._event_id = self.event_id
        v._client = self.client
        self.documents[k] = v

    def __delitem__(self, v: str) -> None:
        raise NotImplementedError()

    def _refresh_documents(self):
        if self.client is not None:
            document_names = self.client.get_all_document_names(self.event_id)
            for name in document_names:
                if name not in self.documents:
                    document = Document(name, event=self.event)
                    self.documents[name] = document


class _Metadata(MutableMapping[str, str]):
    __slots__ = ('_client', '_event', '_event_id', '_metadata')

    def __init__(self, event: Event, client: Optional[EventsClient] = None):
        self._client = client
        self._event = event
        self._event_id = event.event_id
        self._metadata = {}

    def __contains__(self, key):
        if key in self._metadata:
            return True
        self._refresh_metadata()
        return key in self._metadata

    def __setitem__(self, key, value):
        if key in self:
            raise KeyError("Metadata already exists with key: " + key)
        self._metadata[key] = value
        if self._client is not None:
            if not self._client.add_metadata(self._event_id, key, value):
                raise ValueError()

    def __getitem__(self, key):
        try:
            return self._metadata[key]
        except KeyError:
            self._refresh_metadata()
        return self._metadata[key]

    def __delitem__(self, v) -> None:
        raise NotImplementedError

    def __iter__(self) -> Iterator[str]:
        self._refresh_metadata()
        return iter(self._metadata)

    def __len__(self) -> int:
        self._refresh_metadata()
        return len(self._metadata)

    def _refresh_metadata(self):
        if self._client is not None:
            response = self._client.get_all_metadata(self._event_id)
            self._metadata.update(response)


class _Binaries(collections.abc.MutableMapping):
    __slots__ = ('_client', '_event', '_event_id', '_names', '_binaries')

    def __init__(self, event: Event, client: Optional[EventsClient] = None):
        self._client = client
        self._event = event
        self._event_id = event.event_id
        self._names = set()
        self._binaries = {}

    def __contains__(self, key):
        if key in self._names:
            return True
        self._refresh_binaries()
        return key in self._names

    def __setitem__(self, key, value):
        if key in self:
            raise KeyError("Binary already exists with name: " + key)
        self._names.add(key)
        self._binaries[key] = value
        if self._client is not None:
            if not self._client.add_binary_data(self._event_id, key, value):
                raise ValueError()

    def __getitem__(self, key):
        try:
            return self._binaries[key]
        except KeyError:
            pass
        if self._client is not None:
            b = self._client.get_binary_data(event_id=self._event_id, binary_data_name=key)
            self._names.add(b)
            self._binaries[key] = b
        return self._binaries[key]

    def __delitem__(self, v) -> None:
        raise NotImplementedError

    def __iter__(self) -> Iterator[str]:
        self._refresh_binaries()
        return iter(self._names)

    def __len__(self) -> int:
        self._refresh_binaries()
        return len(self._names)

    def _refresh_binaries(self):
        if self._client is not None:
            response = self._client.get_all_binary_data_names(self._event_id)
            self._names.update(response)


class _LabelIndices(Mapping[str, 'data.LabelIndex']):
    __slots__ = ('_document', '_document_name', '_event_id', '_client', '_cache', '_names_cache')

    def __init__(self, document: Document):
        self._document = document
        self._document_name = document.document_name
        try:
            self._event_id = document.event.event_id
            self._client = document.event.client
        except AttributeError:
            self._event_id = None
            self._client = None
        self._cache = {}
        self._names_cache = set()

    def __getitem__(self, k: str) -> 'data.LabelIndex':
        try:
            return self._cache[k]
        except KeyError:
            pass

        if k not in self:
            raise KeyError

        if self._client is not None:
            label_adapter = self._document.get_default_adapter(k)
            if label_adapter is None:
                label_adapter = data.GENERIC_ADAPTER
            index = self._client.get_labels(self._event_id, self._document_name, k,
                                            adapter=label_adapter)
            for label in index:
                label.label_index_name = k
                label.document = self._document
            self._cache[k] = index
            self._names_cache.add(k)
            return index
        else:
            raise KeyError('Document does not have label index: ' + k)

    def __contains__(self, item):
        self._refresh_names()
        return item in self._names_cache

    def __len__(self) -> int:
        self._refresh_names()
        return len(self._names_cache)

    def __iter__(self) -> Iterator[str]:
        self._refresh_names()
        return iter(self._names_cache)

    def _refresh_names(self):
        if self._client is not None:
            infos = self._client.get_label_index_info(self._event_id, self._document_name)
            for info in infos:
                self._names_cache.add(info.index_name)

    def add_to_cache(self, label_index_name, index):
        self._cache[label_index_name] = index
        self._names_cache.add(label_index_name)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import signal
import subprocess

import grpc
import pytest


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "consul"
    )
    config.addinivalue_line(
        "markers", "integration"
    )


def pytest_addoption(parser):
    parser.addoption(
        "--consul", action="store_true", default=False,
        help="runs integration tests that require consul",
    )
    parser.addoption(
        "--integration", action="store_true", default=False, help="runs integration tests"
    )


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--consul"):
        skip_consul = pytest.mark.skip(reason="need --consul option to run")
        for item in items:
            if "consul" in item.keywords:
                item.add_marker(skip_consul)
    if not config.getoption("--integration"):
        skip_integration = pytest.mark.skip("need --integration option to run")
        for item in items:
            if "integration" in item.keywords:
                item.add_marker(skip_integration)


@pytest.fixture(name='processor_watcher')
def fixture_processor_watcher():
    def func(address, process):
        try:
            if process.returncode is not None:
                raise ValueError('subprocess terminated')
            with grpc.insecure_channel(address, [('grpc.enable_http_proxy', False)]) as channel:
                future = grpc.channel_ready_future(channel)
                future.result(timeout=20)
            yield address
        finally:
            process.send_signal(signal.SIGINT)
            try:
                stdout, _ = process.communicate(timeout=1)
                print("processor exited with code: ", process.returncode)
                print(stdout.decode('utf-8'))
            except subprocess.TimeoutExpired:
                print("timed out waiting for processor to terminate")
    return func



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Internal utilities for interacting with protobuf structs."""
from typing import Sequence, Mapping

from google.protobuf import struct_pb2


def copy_dict_to_struct(d, struct, parents=None):
    if parents is None:
        parents = [d]
    for k, v in d.items():
        copy_item_to_struct(k, v, struct, parents)


def copy_item_to_struct(k, v, struct, parents):
    if isinstance(v, (str, float, bool, int)) or v is None:
        struct[k] = v
    elif isinstance(v, Sequence):
        _check_for_reference_cycle(v, parents)
        copy_list_to_list_value(v, struct.get_or_create_list(k), parents + [v])
    elif isinstance(v, Mapping):
        _check_for_reference_cycle(v, parents)
        copy_dict_to_struct(v, struct.get_or_create_struct(k), parents + [v])
    else:
        raise TypeError("Unrecognized type:", type(v))


def copy_list_to_list_value(lst, list_value, parents):
    for v in lst:
        if isinstance(v, (str, float, bool, int)) or v is None:
            list_value.append(v)
        elif isinstance(v, Sequence):
            _check_for_reference_cycle(v, parents)
            copy_list_to_list_value(v, list_value.add_list(), parents + [v])
        elif isinstance(v, Mapping):
            _check_for_reference_cycle(v, parents)
            copy_dict_to_struct(v, list_value.add_struct(), parents + [v])
        else:
            raise TypeError("Unrecognized type:", type(v))


def _check_for_reference_cycle(o, parents):
    for parent in parents:
        if o is parent:
            raise ValueError("Reference cycle while preparing for serialization.")


def copy_list_value_to_list(list_value, lst):
    for v in list_value:
        if isinstance(v, (str, float, bool, int)) or v is None:
            lst.append(v)
        elif isinstance(v, struct_pb2.ListValue):
            lst2 = []
            copy_list_value_to_list(v, lst2)
            lst.append(lst2)
        elif isinstance(v, struct_pb2.Struct):
            d = {}
            copy_struct_to_dict(v, d)
            lst.append(d)
        else:
            raise TypeError("Unrecognized type:", type(v))


def copy_struct_to_dict(struct, dictionary):
    for k, v in struct.items():
        if isinstance(v, (str, float, bool, int)) or v is None:
            dictionary[k] = v
        elif isinstance(v, struct_pb2.ListValue):
            lst = []
            copy_list_value_to_list(v, lst)
            dictionary[k] = lst
        elif isinstance(v, struct_pb2.Struct):
            d = {}
            copy_struct_to_dict(v, d)
            dictionary[k] = d
        else:
            raise TypeError("Unrecognized type:", type(v))




# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Converts brat .ann/.txt pairs into serialized events."""
from argparse import ArgumentParser, Namespace
from pathlib import Path

from mtap.io import brat, serialization


class BratConverter(Namespace):
    """Reads brat documents and writes them to a file using a serializer.

    Attributes
    ----------
    input_dir: str
        The input directory
    output_dir: str
        The output directory.
    serializer: Serializer
        The serializer to use.
    events_address: str
        The address of the events service.
    document_name: str
        The document name to store the BRAT text and annotations.


    """

    def convert_files(self):
        self.output_dir.mkdir(parents=True, exist_ok=True)
        for event in brat.read_brat_documents(self.input_dir, document_name=self.document_name,
                                              label_index_name_prefix=self.label_index_name_prefix,
                                              encoding=self.input_encoding,
                                              create_indices=self.create_indices):
            output_path = self.output_dir.joinpath(event.event_id + self.serializer.extension)
            self.serializer.event_to_file(event, output_path)


def main(args=None):
    parser = ArgumentParser()
    parser.add_argument('input_dir', type=Path,
                        help="The input directory of BRAT .ann and .txt files")
    parser.add_argument('output_dir', type=Path,
                        help="The output directory of serialized files.")
    parser.add_argument('create_indices', nargs='*', default=[],
                        help='Indices to create, even if empty.')
    parser.add_argument('--serializer', type=serialization.get_serializer, default='json')
    parser.add_argument('--events-address', default=None,
                        help='Address of the events service to use.')
    parser.add_argument('--document-name', default='plaintext',
                        help='The document on the event to create and read brat documents into.')
    parser.add_argument('--label-index-name-prefix', default='',
                        help='A string to prefix the brat annotation types before creating label '
                             'indices.')
    parser.add_argument('--input-encoding', default=None,
                        help='The encoding of the input BRAT files.')

    converter = parser.parse_args(args=args, namespace=BratConverter())
    converter.convert_files()


if __name__ == '__main__':
    main()



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Descriptors for processor functionality."""
import typing
from typing import NamedTuple, Optional, List, Type, Dict, Any

if typing.TYPE_CHECKING:
    from mtap import EventProcessor

__all__ = [
    'label_property',
    'labels',
    'parameter',
    'processor',
]


class label_property(NamedTuple('label_property',
                                [('name', str),
                                 ('description', Optional[str]),
                                 ('data_type', Optional[str]),
                                 ('nullable', bool)])):
    """Creates a description for a property on a label.

    Args:
        name (str): The property's name.

    Keyword Args:
        description (~typing.Optional[str]): A short description of the property.
        data_type (~typing.Optional[str]):
            The data type of the property: str, float, or bool; List[T] or Mapping[T1, T2] of those.
        nullable (bool): Whether the property can have a valid value of null.
    """

    def __new__(cls,
                name: str,
                *,
                nullable: bool = False,
                description: Optional[str] = None,
                data_type: Optional[str] = None):
        return super().__new__(cls, name, description, data_type, nullable)


class labels(NamedTuple('LabelDescription',
                        [('name', str),
                         ('reference', Optional[str]),
                         ('name_from_parameter', Optional[str]),
                         ('optional', bool),
                         ('description', Optional[str]),
                         ('properties', List[label_property])])):
    """A description for a label type.

    Args:
        name (str): The label index name.

    Keyword Args:
        reference (~typing.Optional[str]):
            If this is an output of another processor, that processor's name followed by a slash
            and the default output name of the index go here.
            Example: "sentence-detector/sentences".
        optional (bool): Whether this label index is an optional input or output.
        name_from_parameter (~typing.Optional[str]):
            If the label index gets its name from a processor parameter, the name of the parameter.
        description (~typing.Optional[str]): A short description of the label index.
        properties (~typing.Optional[~typing.List[label_property]]):
            The properties of the labels in the label index.

    Attributes:
        name (str): The label index name.
        reference (~typing.Optional[str]):
            If this is an output of another processor, that processor's name followed by a slash
            and the default output name of the index go here.
            Example: "sentence-detector/sentences".
        optional (bool): Whether this label index is an optional input or output.
        name_from_parameter (~typing.Optional[str]):
            If the label index gets its name from a processor parameter, the name of the parameter.
        description (~typing.Optional[str]): A short description of the label index.
        properties (~typing.Optional[~typing.List[label_property]]):
            The properties of the labels in the label index.
    """

    def __new__(cls,
                name: str,
                *,
                reference: Optional[str] = None,
                optional: bool = False,
                name_from_parameter: Optional[str] = None,
                description: Optional[str] = None,
                properties: Optional[List[label_property]] = None):
        return super().__new__(cls, name, reference, name_from_parameter, optional, description,
                               properties)


class parameter(NamedTuple('parameter',
                           [('name', str),
                            ('description', Optional[str]),
                            ('data_type', Optional[str]),
                            ('required', bool)])):
    def __new__(cls,
                name: str,
                *,
                required: bool = False,
                data_type: Optional[str] = None,
                description: Optional[str] = None):
        """A description of one of the processor's parameters.

        Args:
            name (str): The parameter name / key.

        Keyword Args:
            required (bool): Whether the processor parameter is required.
            data_type (~typing.Optional[str]):
                The data type of the parameter. str, float, or bool; List[T] or Mapping[T1, T2] of
                those.
            description (~typing.Optional[str]): A short description of the property and what it
                does.

        Attributes:
            name (str): The parameter name / key.
            required (bool): Whether the processor parameter is required.
            data_type (~typing.Optional[str]):
                The data type of the parameter. str, float, or bool; List[T] or Mapping[T1, T2] of
                those.
            description (~typing.Optional[str]): A short description of the property and what it
                does.


        Returns:
            ParameterDescription: The parameter description object
        """
        return super().__new__(cls, name=name, description=description, data_type=data_type,
                               required=required)


def processor(name: str,
              *,
              human_name: Optional[str] = None,
              description: Optional[str] = None,
              parameters: Optional[List[parameter]] = None,
              inputs: Optional[List[labels]] = None,
              outputs: Optional[List[labels]] = None,
              **additional_metadata: str):
    """Decorator which attaches a service name and metadata to a processor. Which then can be used
    for runtime reflection of how the processor works.

    Args:
        name (str):
            Identifying service name both for launching via command line and for service
            registration.

            Should be a mix of alphanumeric characters and dashes so that it plays nice with the
            DNS name requirements of service discovery tools like Consul. Can be overridden at
            runtime via the `identifier` option on :func:`processor_parser`.

    Keyword Args:
        human_name (~typing.Optional[str]): An option human name for the processor.
        description (~typing.Optional[str]): A short description of the processor and what it does.
        parameters (~typing.Optional[~typing.List[ParameterDescription]]):
            The processor's parameters.
        inputs (~typing.Optional[~typing.List[str]]):
            String identifiers for the output from a processor that this processor uses as an input.

            Takes the format "[processor-name]/[output]". Examples would be "tagger:pos_tags" or
            "sentence-detector:sentences".
        outputs (~typing.Optional[~typing.List[LabelDescription]]):
            The label indices this processor outputs.
        **additional_metadata (~typing.Any):
            Any other data that should be added to the processor's metadata, should be serializable
            to yaml and json.

    Returns:
        A decorator to be applied to instances of EventProcessor or DocumentProcessor. This
        decorator attaches the metadata so it can be reflected at runtime.

    Examples:
        >>> from mtap.processing import EventProcessor
        >>> @processor('example-text-converter')
        >>> class TextConverter(EventProcessor):
        >>>     ...

        or

        >>> from mtap.processing import DocumentProcessor
        >>> @processor('example-sentence-detector')
        >>> class SentenceDetector(DocumentProcessor):
        >>>     ...

        From our own example processor:

        >>> from mtap.processing import DocumentProcessor
        >>> @processor('mtap-example-processor-python',
        >>>            human_name="Python Example Processor",
        >>>            description="counts the number of times the letters a and b occur in a document",
        >>>            parameters=[
        >>>                parameter('do_work', required=True, data_type='bool',
        >>>                          description="Whether the processor should do anything.")
        >>>            ],
        >>>            outputs=[
        >>>                labels('mtap.examples.letter_counts',
        >>>                            properties=[label_property('letter', data_type='str'),
        >>>                                        label_property('count', data_type='int')])
        >>>            ])
        >>> class ExampleProcessor(DocumentProcessor):
        >>>     ...


    """

    def decorator(f: Type['EventProcessor']) -> Type['EventProcessor']:
        f.metadata = {
            'name': name,
            'human_name': human_name,
            'description': description
        }
        if parameters is not None:
            f.metadata['parameters'] = [_parameter_to_map(p) for p in parameters]
        if inputs is not None:
            f.metadata['inputs'] = [_desc_to_dict(desc) for desc in inputs]
        if outputs is not None:
            f.metadata['outputs'] = [_desc_to_dict(desc) for desc in outputs]
        if additional_metadata is not None:
            f.metadata.update(additional_metadata)
        if 'implementation_lang' not in f.metadata:
            f.metadata['implementation_lang'] = 'Python'
        return f

    return decorator


def _desc_to_dict(description: labels) -> dict:
    m = {
        'name': description.name
    }
    if description.name_from_parameter is not None:
        m['name_from_parameter'] = description.name_from_parameter
    if description.reference is not None:
        m['reference'] = description.reference
    if description.optional is not None:
        m['optional'] = description.optional
    if description.description is not None:
        m['description'] = description.description
    if description.properties is not None:
        m['properties'] = [_prop_to_dict(p) for p in description.properties]
    return m


def _prop_to_dict(p: label_property) -> Dict[str, Any]:
    m = {
        'name': p.name
    }
    if p.description is not None:
        m['description'] = p.description
    if p.data_type is not None:
        m['data_type'] = p.data_type
    if p.nullable is not None:
        m['nullable'] = p.nullable
    return m


def _parameter_to_map(p: parameter) -> Dict[str, Any]:
    m = {
        'name': p.name
    }
    if p.description is not None:
        m['description'] = p.description
    if p.data_type is not None:
        m['data_type'] = p.data_type
    if p.required is not None:
        m['required'] = p.required
    return m



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Prints the processor metadata for the example processor."""
import sys
from os import environ
from subprocess import call
from tempfile import NamedTemporaryFile

import mtap.examples.example_processor as example_processor


def main(filename: str):
    java_jar = environ['MTAP_JAR']

    from yaml import dump, load
    try:
        from yaml import CDumper as Dumper, CLoader as Loader
    except ImportError:
        from yaml import Dumper as Dumper, Loader as Loader

    with NamedTemporaryFile('r') as f:
        # Here we call out to a java utility for printing processor metadata and then load that
        # metadata to merge it with the python metadata.
        # The classpath should include MTAP and class files for any processors for which you want
        # to print metadata.
        call(['java', '-cp', java_jar, 'edu.umn.nlpie.mtap.utilities.PrintProcessorMetadata',
              f.name,
              'edu.umn.nlpie.mtap.examples.WordOccurrencesExampleProcessor',
              ])
        java_meta = load(f, Loader=Loader)

    with open(filename, 'w') as f:
        dump([example_processor.ExampleProcessor.metadata] + java_meta, f, Dumper=Dumper)


if __name__ == '__main__':
    main(sys.argv[1])



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from mtap.api.v1 import processing_pb2 as mtap_dot_api_dot_v1_dot_processing__pb2


class ProcessorStub(object):
    """A processor service is responsible for processing an event, adding results to
    the events service, and then optionally returning a custom result.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Process = channel.unary_unary(
                '/mtap.api.v1.Processor/Process',
                request_serializer=mtap_dot_api_dot_v1_dot_processing__pb2.ProcessRequest.SerializeToString,
                response_deserializer=mtap_dot_api_dot_v1_dot_processing__pb2.ProcessResponse.FromString,
                )
        self.GetInfo = channel.unary_unary(
                '/mtap.api.v1.Processor/GetInfo',
                request_serializer=mtap_dot_api_dot_v1_dot_processing__pb2.GetInfoRequest.SerializeToString,
                response_deserializer=mtap_dot_api_dot_v1_dot_processing__pb2.GetInfoResponse.FromString,
                )
        self.GetStats = channel.unary_unary(
                '/mtap.api.v1.Processor/GetStats',
                request_serializer=mtap_dot_api_dot_v1_dot_processing__pb2.GetStatsRequest.SerializeToString,
                response_deserializer=mtap_dot_api_dot_v1_dot_processing__pb2.GetStatsResponse.FromString,
                )


class ProcessorServicer(object):
    """A processor service is responsible for processing an event, adding results to
    the events service, and then optionally returning a custom result.
    """

    def Process(self, request, context):
        """Processes an event.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetInfo(self, request, context):
        """Gets information about the processor.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStats(self, request, context):
        """Gets globally-aggregated statistics about the processor.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ProcessorServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Process': grpc.unary_unary_rpc_method_handler(
                    servicer.Process,
                    request_deserializer=mtap_dot_api_dot_v1_dot_processing__pb2.ProcessRequest.FromString,
                    response_serializer=mtap_dot_api_dot_v1_dot_processing__pb2.ProcessResponse.SerializeToString,
            ),
            'GetInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetInfo,
                    request_deserializer=mtap_dot_api_dot_v1_dot_processing__pb2.GetInfoRequest.FromString,
                    response_serializer=mtap_dot_api_dot_v1_dot_processing__pb2.GetInfoResponse.SerializeToString,
            ),
            'GetStats': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStats,
                    request_deserializer=mtap_dot_api_dot_v1_dot_processing__pb2.GetStatsRequest.FromString,
                    response_serializer=mtap_dot_api_dot_v1_dot_processing__pb2.GetStatsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'mtap.api.v1.Processor', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Processor(object):
    """A processor service is responsible for processing an event, adding results to
    the events service, and then optionally returning a custom result.
    """

    @staticmethod
    def Process(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/mtap.api.v1.Processor/Process',
            mtap_dot_api_dot_v1_dot_processing__pb2.ProcessRequest.SerializeToString,
            mtap_dot_api_dot_v1_dot_processing__pb2.ProcessResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/mtap.api.v1.Processor/GetInfo',
            mtap_dot_api_dot_v1_dot_processing__pb2.GetInfoRequest.SerializeToString,
            mtap_dot_api_dot_v1_dot_processing__pb2.GetInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/mtap.api.v1.Processor/GetStats',
            mtap_dot_api_dot_v1_dot_processing__pb2.GetStatsRequest.SerializeToString,
            mtap_dot_api_dot_v1_dot_processing__pb2.GetStatsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#  Copyright 2021 Regents of the University of Minnesota.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pickle
import time
from datetime import timedelta
from pathlib import Path
from typing import Dict, Any, Callable, Union

import pytest

from mtap import Pipeline, LocalProcessor, Event, EventsClient, Document
from mtap.processing import EventProcessor, ProcessingError, PipelineResult, ProcessingSource
from mtap.processing._pipeline import MpConfig, RemoteProcessor


class Processor(EventProcessor):
    def __init__(self, identifier='1'):
        self.identifier = identifier
        self.seen = 0
        self.processed = 0

    def process(self, event: Event, params: Dict[str, Any]):
        self.seen += 1
        if 'fail' in event.event_id:
            raise ValueError("fail")
        time.sleep(0.001)
        event.metadata[self.identifier] = 'True'
        event.metadata['processor'] = self.identifier
        self.processed += 1


def test_time_result(mocker):
    client = mocker.Mock(EventsClient)
    client.get_local_instance.return_value = client
    client.get_all_document_names.return_value = ['plaintext']
    client.get_all_metadata.return_value = {}
    client.instance_id = 0
    with Pipeline(
            LocalProcessor(Processor(), component_id='test_processor'),
            events_client=client
    ) as pipeline:
        event = Event()
        result = pipeline.run(event)
        assert result.component_results[0].timing_info['process_method'] >= timedelta(seconds=0.001)



def test_load_from_config():
    pipeline = Pipeline.from_yaml_file(Path(__file__).parent / 'pipeline.yml')
    assert pipeline.name == 'mtap-test-pipeline'
    assert pipeline.events_address == 'localhost:123'
    assert pipeline.mp_config.max_failures == 3
    assert not pipeline.mp_config.show_progress
    assert pipeline.mp_config.workers == 12
    assert pipeline.mp_config.read_ahead == 4
    assert not pipeline.mp_config.close_events
    assert len(pipeline) == 2
    assert pipeline[0].processor_id == 'processor-1'
    assert pipeline[0].address == 'localhost:1234'
    assert pipeline[1].processor_id == 'processor-2'
    assert pipeline[1].address == 'localhost:5678'


def test_serialization():
    p = Pipeline(
        RemoteProcessor(
            processor_id='processor-1',
            address='localhost:1234'
        ),
        RemoteProcessor(
            processor_id='processor-2',
            address='localhost:5678'
        ),
        name='mtap-test-pipeline',
        events_address='localhost:123',
        mp_config=MpConfig(
            max_failures=3,
            show_progress=False,
            workers=12,
            read_ahead=4,
            close_events=False
        ),
    )
    s = pickle.dumps(p)
    r = pickle.loads(s)
    assert r.name == 'mtap-test-pipeline'
    assert r.events_address == 'localhost:123'
    assert r.mp_config.max_failures == 3
    assert not r.mp_config.show_progress
    assert r.mp_config.workers == 12
    assert r.mp_config.read_ahead == 4
    assert not r.mp_config.close_events
    assert len(r) == 2
    assert r[0].processor_id == 'processor-1'
    assert r[0].address == 'localhost:1234'
    assert r[1].processor_id == 'processor-2'
    assert r[1].address == 'localhost:5678'



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tools for reading in files annotated using BRAT (https://brat.nlplab.org/)."""

from contextlib import contextmanager
from pathlib import Path
from typing import Union, Iterable, Optional, TYPE_CHECKING

from mtap import data

if TYPE_CHECKING:
    import mtap

__all__ = [
    'ann_to_labels',
    'read_brat_document',
    'read_brat_documents',
]


def read_brat_documents(directory: Union[Path, str],
                        client: Optional['mtap.EventsClient'] = None,
                        document_name: str = 'plaintext',
                        label_index_name_prefix: str = '',
                        encoding: Optional[str] = None,
                        create_indices: Optional[Iterable[str]] = None) -> Iterable['mtap.Event']:
    """Reads a directory full of BRAT-annotated ".txt" and ".ann" files, creating events for each
    .txt containing labels created from the annotations in its ".ann" counterpart.

    Args:
        directory (str, Path): The directory containing the files or sub-directories containing the
            files.
        client (EventsClient): The client to the events server to add the documents to.
        document_name (str): The document name to create on each event containing the brat data.
        label_index_name_prefix  (str): A prefix to append to the brat annotation names.
        encoding (optional str): The encoding to use when reading the files.
        create_indices (optional iterable of str): These indices will be created no matter what,
            even if empty.

    Returns:
        Iterable of Event: Iterable of all the events from the directory of brat documents.

    Examples:
        >>> client = mtap.EventsClient()
        >>> for event in read_brat_documents('docs', client=client):
        >>>   with event:
        >>>      # use event

    """
    directory = Path(directory)
    for txt_file in directory.glob('**/*.txt'):
        with read_brat_document(txt_file, client=client, document_name=document_name,
                                label_index_name_prefix=label_index_name_prefix, encoding=encoding,
                                relative_to=directory, create_indices=create_indices) as event:
            yield event


@contextmanager
def read_brat_document(txt_file: Union[Path, str],
                       client: Optional['mtap.EventsClient'] = None,
                       document_name: str = 'plaintext',
                       label_index_name_prefix: str = '',
                       encoding: Optional[str] = None,
                       relative_to: Optional[Union[Path, str]] = None,
                       create_indices: Optional[Iterable[str]] = None) -> 'mtap.Event':
    """Reads a BRAT .txt and .ann file pair into an Event.

    Args:
        txt_file (Path or str): The path to the .txt file.
        client (optional EventsClient): The client to the events server to add the documents to.
        document_name (str): The document name to create on the event to hold the brat data.
        label_index_name_prefix (str): A prefix to append to the brat annotation names.
        encoding (optional str): The encoding to use when reading the files.
        relative_to (optional str or Path): A str of a path or a Path object for which the event_id
            will be relative to, if this is omitted then the file's name without extension will be
            used.
        create_indices (optional iterable of str): These indices will be created no matter what,
            even if empty.

    Returns:
        Event: An event containing the text and annotations as labels.

    """
    txt_file = Path(txt_file)
    ann_file = txt_file.with_suffix('.ann')
    if relative_to is not None:
        relative_to = Path(relative_to)
        event_id = str(txt_file.with_suffix('').relative_to(relative_to))
    else:
        event_id = txt_file.stem
    with data.Event(event_id=event_id, client=client) as event:
        with txt_file.open('r', encoding=encoding) as f:
            txt = f.read()
        document = event.create_document(document_name=document_name, text=txt)
        ann_to_labels(ann_file, document, label_index_name_prefix=label_index_name_prefix,
                      encoding=encoding, create_indices=create_indices)
        yield event


def ann_to_labels(ann_file: Union[str, Path],
                  document: 'mtap.Document',
                  label_index_name_prefix: str,
                  encoding: Optional[str],
                  create_indices: Optional[Iterable[str]] = None):
    """Reads all of the annotations in a brat annotations file into a document.

    Args:
        ann_file (str or Path): A BRAT .ann file to load annotations from.
        document (Document): The document to add labels to.
        label_index_name_prefix (str): A prefix to append to the brat annotation names.
        encoding (optional str): The encoding to use when reading the ann file.
        create_indices (optional iterable of str): These indices will be created no matter what,
            even if empty.
    """
    labelers = {}
    if create_indices is not None:
        for index in create_indices:
            labelers[index] = document.get_labeler(index)
    ann_file = Path(ann_file)
    with ann_file.open('r', encoding=encoding) as f:
        for line in f.readlines():
            splits = line.split('\t')
            if len(splits) < 3 or not splits[0].startswith('T'):
                continue
            name, bounds = splits[1].split(' ', maxsplit=1)
            name = label_index_name_prefix + name
            bounds = bounds.split(';')
            min_start = float('Inf')
            max_end = 0
            for pair in bounds:
                start_index, end_index = pair.split(' ')
                min_start = min(min_start, int(start_index))
                max_end = max(max_end, int(end_index))
            try:
                labeler = labelers[name]
            except KeyError:
                labeler = document.get_labeler(name)
                labelers[name] = labeler
            labeler(min_start, max_end)
    for labeler in labelers.values():
        labeler.done()



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import os
from pathlib import Path

import pytest

from mtap import Config

NO_KEY = object()


@contextlib.contextmanager
def set_env(key, val):
    old = NO_KEY
    try:
        old = os.environ[key]
    except KeyError:
        pass
    os.environ[key] = val
    yield
    if old is NO_KEY:
        del os.environ[key]
    else:
        os.environ[key] = old


def test_load_broken_config():
    Config._global_instance = None
    with set_env('MTAP_CONFIG', str(Path(__file__).parent / 'brokenConfig.yaml')):
        with pytest.raises(TypeError):
            Config()


def test_load_config():
    with set_env('MTAP_CONFIG', str(Path(__file__).parent / 'workingConfig.yaml')):
        Config._global_instance = None
        c = Config()
        assert c['foo'] == 'bar'
        assert c['baz.bot'] == [1, 2, 3]


def test_config_context():
    Config._global_instance = None
    with Config() as c1:
        c1['foo'] = 'bar'
        c2 = Config()
        assert c2['foo'] == 'bar'


def test_enter_twice():
    Config._global_instance = None
    with Config():
        with pytest.raises(ValueError):
            with Config():
                pass


def test_update_from_yaml():
    Config._global_instance = None
    c = Config()
    c.update_from_yaml(str(Path(__file__).parent) + '/workingConfig.yaml')
    assert c['foo'] == 'bar'
    assert c['baz.bot'] == [1, 2, 3]



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Internal events service client"""

import logging
import threading
import typing
import uuid
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Optional

import grpc
from google.protobuf import empty_pb2
from grpc_health.v1 import health
from grpc_health.v1 import health_pb2_grpc

from mtap import _config, constants, utilities
from mtap.api.v1 import events_pb2, events_pb2_grpc

if typing.TYPE_CHECKING:
    import mtap

LOGGER = logging.getLogger(__name__)


class EventsServer:
    """Server which hosts events.

    Args:
        host (str): The address / hostname / IP to host the server on.

    Keyword Args:
        port (int): The port to host the server on.
        register (bool): Whether to register the service with service discovery.
        workers (int): The number of workers that should handle requests.
        write_address (bool): Whether to write the events service address to a file.
        config (mtap.Config): An optional mtap config.
    """

    def __init__(self, host: str, *, port: int = 0, register: bool = False, workers: int = 10,
                 write_address: bool = False, config: 'Optional[mtap.Config]' = None):
        if host is None:
            host = '127.0.0.1'
        if port is None:
            port = 0
        self.write_address = write_address
        thread_pool = ThreadPoolExecutor(max_workers=workers)
        if config is None:
            config = _config.Config()
        options = config.get('grpc.events_server_options', {})
        server = grpc.server(thread_pool, options=list(options.items()))
        servicer = EventsServicer()
        events_pb2_grpc.add_EventsServicer_to_server(servicer, server)
        health_servicer = health.HealthServicer()
        health_servicer.set('', 'SERVING')
        health_servicer.set(constants.EVENTS_SERVICE_NAME, 'SERVING')
        health_pb2_grpc.add_HealthServicer_to_server(health_servicer, server)
        self._port = server.add_insecure_port(host + ':' + str(port))
        self._server = server
        self._address = host
        self._config = _config.Config()
        self._register = register
        self._address_file = None

    @property
    def port(self) -> int:
        """int: The port the processor service is bound to."""
        return self._port

    def start(self):
        """Starts the service.
        """
        LOGGER.info('Starting events server on address: "%s:%d"', self._address, self._port)
        self._server.start()
        if self.write_address:
            self._address_file = utilities.write_address_file(
                '{}:{}'.format(self._address, self.port)
            )
        if self._register:
            from mtap._discovery import Discovery
            service_registration = Discovery(config=self._config)
            self._deregister = service_registration.register_events_service(self._address,
                                                                            self._port,
                                                                            'v1')

    def stop(self, *, grace: Optional[float] = None):
        """De-registers (if registered with service discovery) the service and immediately stops
        accepting requests, completely stopping the service after a specified `grace` time.

        During the grace period the server will continue to process existing requests, but it will
        not accept any new requests. This function is idempotent, multiple calls will shutdown
        the server after the soonest grace to expire, calling the shutdown event for all calls to
        this function.

        Keyword Args:
            grace (~typing.Optional[float]):
                The grace period in seconds that the server should continue processing requests
                before shutdown.

        Returns:
            threading.Event: A shutdown event for the server.
        """
        print("Shutting down events server on address: {}:{}".format(self._address, self._port))
        if self._address_file is not None:
            self._address_file.unlink()
        try:
            self._deregister()
        except AttributeError:
            pass
        return self._server.stop(grace=grace)


class EventsServicer(events_pb2_grpc.EventsServicer):
    def __init__(self):
        self.lock = threading.RLock()
        self.events = {}
        self.instance_id = str(uuid.uuid4())

    def _get_event(self, request, context=None):
        event_id = request.event_id
        try:
            event = self.events[event_id]
        except KeyError as e:
            _set_error_context(context, grpc.StatusCode.NOT_FOUND,
                               "Did not find event_id: '{}'".format(event_id))
            raise e
        return event, event_id

    def _get_document(self, request, context=None):
        event, event_id = self._get_event(request, context)
        document_name = request.document_name
        try:
            document = event.documents[document_name]
        except KeyError as e:
            _set_error_context(context, grpc.StatusCode.NOT_FOUND,
                               "Event: '{}' does not have document: '{}'".format(event_id,
                                                                                 document_name))
            raise e
        return document

    def GetEventsInstanceId(self, request, context):
        LOGGER.debug("GetEventsInstanceId")
        return events_pb2.GetEventsInstanceIdResponse(instance_id=self.instance_id)

    def OpenEvent(self, request, context=None):
        LOGGER.debug("OpenEvent: %s", request.event_id)
        event_id = request.event_id
        if event_id == '':
            msg = "event_id was not set."
            _set_error_context(context, grpc.StatusCode.INVALID_ARGUMENT, msg)
            return empty_pb2.Empty()
        created_event = False
        try:
            event = self.events[event_id]
        except KeyError:
            with self.lock:
                try:
                    event = self.events[event_id]
                except KeyError:
                    created_event = True
                    event = _Event()
                    self.events[event_id] = event

        if not created_event and request.only_create_new:
            msg = 'Event already exists: "{}"'.format(event_id)
            _set_error_context(context, grpc.StatusCode.ALREADY_EXISTS, msg)
            return empty_pb2.Empty()
        event.clients += 1
        return events_pb2.OpenEventResponse(created=created_event)

    def CloseEvent(self, request, context=None):
        LOGGER.debug("CloseEvent: %s", request.event_id)
        try:
            event, event_id = self._get_event(request, context)
        except KeyError:
            return empty_pb2.Empty()
        deleted = False
        with event.c_lock:
            event.clients -= 1
            if event.clients == 0:
                del self.events[event_id]
                deleted = True
        return events_pb2.CloseEventResponse(deleted=deleted)

    def GetAllMetadata(self, request, context=None):
        event, _ = self._get_event(request, context)
        return events_pb2.GetAllMetadataResponse(metadata=event.metadata)

    def AddMetadata(self, request, context=None):
        try:
            event, _ = self._get_event(request, context)
        except KeyError:
            return empty_pb2.Empty()
        key = request.key
        if key == '':
            msg = 'metadata key cannot be null or empty'
            _set_error_context(context, grpc.StatusCode.INVALID_ARGUMENT, msg)
            return empty_pb2.Empty()
        event.metadata[key] = request.value
        return events_pb2.AddMetadataResponse()

    def GetAllBinaryDataNames(self, request, context):
        try:
            event, event_id = self._get_event(request, context)
        except KeyError:
            return empty_pb2.Empty()
        names = list(event.binaries.keys())
        return events_pb2.GetAllBinaryDataNamesResponse(binary_data_names=names)

    def AddBinaryData(self, request, context):
        try:
            event, event_id = self._get_event(request, context)
        except KeyError:
            return empty_pb2.Empty()
        name = request.binary_data_name
        if name == '':
            msg = 'binary_data_name cannot be null or empty'
            _set_error_context(context, grpc.StatusCode.INVALID_ARGUMENT, msg)
            return empty_pb2.Empty()
        event.binaries[name] = request.binary_data
        return events_pb2.AddBinaryDataResponse()

    def GetBinaryData(self, request, context):
        try:
            event, event_id = self._get_event(request, context)
        except KeyError:
            return empty_pb2.Empty()
        name = request.binary_data_name
        if name == '':
            msg = 'binary_data_name cannot be null or empty'
            _set_error_context(context, grpc.StatusCode.INVALID_ARGUMENT, msg)
            return empty_pb2.Empty()
        return events_pb2.GetBinaryDataResponse(binary_data=event.binaries[name])

    def AddDocument(self, request, context=None):
        LOGGER.debug("AddDocument: %s", request.event_id)
        try:
            event, event_id = self._get_event(request, context)
        except KeyError:
            return empty_pb2.Empty()
        document_name = request.document_name
        if document_name == '':
            msg = 'document_name was not set.'
            _set_error_context(context, grpc.StatusCode.INVALID_ARGUMENT, msg)
            return empty_pb2.Empty()

        with event.d_lock:
            if document_name in event.documents:
                msg = "Document '{}' already exists.".format(document_name)
                _set_error_context(context, grpc.StatusCode.ALREADY_EXISTS, msg)
                return empty_pb2.Empty()
            event.documents[document_name] = _Document(request.text)

        return events_pb2.AddDocumentResponse()

    def GetAllDocumentNames(self, request, context=None):
        LOGGER.debug("GetAllDocumentNames: %s", request.event_id)
        try:
            event, event_id = self._get_event(request, context)
        except KeyError:
            return empty_pb2.Empty()
        names = list(event.documents.keys())
        return events_pb2.GetAllDocumentNamesResponse(document_names=names)

    def GetDocumentText(self, request, context=None):
        LOGGER.debug("GetDocumentText: %s", request.event_id)
        try:
            document = self._get_document(request, context)
        except KeyError:
            return empty_pb2.Empty()
        return events_pb2.GetDocumentTextResponse(text=document.text)

    def GetLabelIndicesInfo(self, request, context=None):
        try:
            document = self._get_document(request, context)
        except KeyError:
            return empty_pb2.Empty()
        response = events_pb2.GetLabelIndicesInfoResponse()
        for k, (labels_type, _) in document.labels.items():
            info = response.label_index_infos.add()
            info.index_name = k
            if labels_type == 'generic_labels':
                info.type = events_pb2.GetLabelIndicesInfoResponse.LabelIndexInfo.GENERIC
            elif labels_type == 'custom_labels':
                info.type = events_pb2.GetLabelIndicesInfoResponse.LabelIndexInfo.CUSTOM
        return response

    def AddLabels(self, request, context=None):
        try:
            document = self._get_document(request, context)
        except KeyError:
            return empty_pb2.Empty()
        labels_field = request.WhichOneof('labels')
        if labels_field is None:
            labels = ('generic_labels', events_pb2.GenericLabels())
        else:
            index_name = request.index_name
            if index_name == '':
                msg = 'No index_name was set.'
                _set_error_context(context, grpc.StatusCode.INVALID_ARGUMENT, msg)
                return empty_pb2.Empty()
            labels = (labels_field, getattr(request, labels_field))
        if labels_field == 'generic_labels' and not request.no_key_validation:
            for label in labels[1].labels:
                for key in label.fields:
                    if key in ["document", "location", "text", "id", "label_index_name"]:
                        _set_error_context(context, grpc.StatusCode.INVALID_ARGUMENT,
                                           "Label included a reserved key: {}".format(key))
                        return empty_pb2.Empty()
                for key in label.reference_ids:
                    if key in ["document", "location", "text", "id", "id", "label_index_name"]:
                        _set_error_context(context, grpc.StatusCode.INVALID_ARGUMENT,
                                           "Label included a reserved key: {}".format(key))
                        return empty_pb2.Empty()
        document.labels[request.index_name] = labels
        return events_pb2.AddLabelsResponse()

    def GetLabels(self, request, context=None):
        try:
            document = self._get_document(request, context)
        except KeyError:
            return empty_pb2.Empty()
        try:
            labels_type, labels = document.labels[request.index_name]
        except KeyError:
            msg = "Event: '{}' document: '{} does not have label index: {}'".format(
                request.event_id, request.document_name, request.index_name)
            _set_error_context(context, grpc.StatusCode.NOT_FOUND, msg)
            return empty_pb2.Empty()
        response = events_pb2.GetLabelsResponse()
        if labels_type is not None:
            getattr(response, labels_type).CopyFrom(labels)
        return response


class _Event:
    def __init__(self):
        self.c_lock = threading.RLock()
        self.clients = 0
        self.metadata = {}
        self.documents = {}
        self.binaries = {}
        self.d_lock = threading.RLock()


class _Document:
    def __init__(self, text):
        self.text = text
        self.lock = {}
        self.labels = {}


def _set_error_context(context, status_code, msg):
    try:
        context.set_code(status_code)
        context.set_details(msg)
    except AttributeError:
        pass



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Optional, Dict, Any, TYPE_CHECKING

import grpc

from mtap import _discovery, _structs, Config
from mtap import data
from mtap.api.v1 import processing_pb2, processing_pb2_grpc
from mtap.processing import _base

if TYPE_CHECKING:
    import mtap

logger = logging.getLogger(__name__)


class ProcessorRunner(_base.ProcessingComponent):
    __slots__ = ('processor', 'component_id', 'params', 'metadata', 'processor_id', 'client')

    def __init__(self,
                 proc: 'mtap.EventProcessor',
                 client: 'Optional[mtap.EventsClient]',
                 identifier: Optional[str] = None,
                 params: Optional[Dict[str, Any]] = None):
        self.processor = proc
        self.params = params or {}
        self.metadata = proc.metadata
        self.processor_id = identifier
        self.component_id = identifier
        self.client = client

    def call_process(self, event_id, event_instance_id, params):
        p = dict(self.params)
        if params is not None:
            p.update(params)

        with _base.Processor.enter_context() as c, \
                data.Event(event_id=event_id,
                           event_service_instance_id=event_instance_id,
                           client=self.client) as event:
            with _base.Processor.started_stopwatch('process_method'):
                try:
                    result = self.processor.process(event, p)
                except Exception as e:
                    logger.error(
                        "Error while processing event_id %s through pipeline component  '%s'",
                        event_id, self.component_id)
                    raise e
            return result, c.times, event.created_indices

    def close(self):
        pass


class RemoteRunner(_base.ProcessingComponent):
    __slots__ = ('processor_id', 'component_id', '_address', 'params', '_channel', '_stub')

    def __init__(self, processor_id, component_id, address=None, params=None,
                 enable_proxy=False):
        self.processor_id = processor_id
        self.component_id = component_id
        self._address = address
        self.params = params
        address = self._address
        config = Config()
        if address is None:
            discovery = _discovery.Discovery(config)
            address = discovery.discover_processor_service(processor_id, 'v1')
        channel_options = config.get('grpc.processor_channel_options', {})
        self._channel = grpc.insecure_channel(address, options=list(channel_options.items()))
        self._stub = processing_pb2_grpc.ProcessorStub(self._channel)

    def call_process(self, event_id, event_instance_id, params):
        logger.debug('processor_id: %s calling process on event_id: %s with event_instance_id: %s',
                     self.processor_id, event_id, event_instance_id)
        p = dict(self.params or {})
        if params is not None:
            p.update(params)

        with _base.Processor.enter_context() as context:
            request = processing_pb2.ProcessRequest(processor_id=self.processor_id,
                                                    event_id=event_id,
                                                    event_service_instance_id=event_instance_id)
            _structs.copy_dict_to_struct(p, request.params, [p])
            with _base.Processor.started_stopwatch('remote_call'):
                try:
                    response = self._stub.Process(request)
                except Exception as e:
                    logger.error("Error while processing event_id %s through remote pipeline "
                                 "component  '%s'", event_id, self.component_id)
                    raise e
            r = {}
            _structs.copy_struct_to_dict(response.result, r)

            timing_info = response.timing_info
            for k, v in timing_info.items():
                context.add_time(k, v.ToTimedelta())

            created_indices = {}
            for created_index in response.created_indices:
                try:
                    doc_created_indices = created_indices[created_index.document_name]
                except KeyError:
                    doc_created_indices = []
                    created_indices[created_index.document_name] = doc_created_indices
                doc_created_indices.append(created_index.index_name)

            return r, context.times, created_indices

    def close(self):
        self._channel.close()



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides functionality for measuring processor performance against gold standards."""
from abc import ABC, abstractmethod
from typing import Dict, Any, Optional, Sequence, NamedTuple, Callable, Tuple, TextIO, TYPE_CHECKING

import sys

__all__ = [
    'Metric',
    'Metrics',
    'print_overlapping',
    'fields_match_test',
    'Accuracy',
    'ConfusionMatrix',
    'FirstTokenConfusion'
]

from mtap.data import presorted_label_index
from mtap.processing import DocumentProcessor, processor
from mtap.utilities import tokenization

if TYPE_CHECKING:
    import mtap
    from mtap import data, label_index


class Metric(ABC):
    """Base class for metrics.

    """
    name = None

    @abstractmethod
    def update(self,
               document: 'mtap.Document',
               tested_index: 'data.LabelIndex', target_index: 'data.LabelIndex') -> Any:
        ...


@processor('mtap-metrics')
class Metrics(DocumentProcessor):
    """A document process that computes a set of metrics.

    Args
    ---
    tested (str): The name of the index to use as the hypothesis / predictions.
    target (str): The name of the index to use as the ground truth / gold standard.
    tested_filter (~typing.Callable[[Label], bool]): A filter to apply to the tested index.
    target_filter (~typing.Callable[[Label], bool]): A filter to apply to the target index.

    """

    def __init__(self, *metrics: Metric,
                 tested: str,
                 target: str,
                 tested_filter: Callable[['data.Label'], bool] = None,
                 target_filter: Callable[['data.Label'], bool] = None):
        self.tested = tested
        self.target = target
        self.metrics = metrics
        self.tested_filter = tested_filter
        self.target_filter = target_filter

    def process_document(self, document: 'mtap.Document',
                         params: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        tested = document.labels[self.tested]
        if self.tested_filter is not None:
            tested = tested.filter(self.tested_filter)
        target = document.labels.get(self.target, presorted_label_index([]))
        if self.target_filter is not None:
            target = target.filter(self.target_filter)
        local = {}
        for metric in self.metrics:
            local[metric.name] = metric.update(document, tested, target)
        return local


def print_overlapping(document, target_label, tested_index):
    print('Not found:', target_label)
    print('"', target_label.text(document.text), '"')
    overlapping = tested_index.overlapping(target_label)
    print('Overlapping:', overlapping)
    for overlap in overlapping:
        print('"', overlap.text(document.text), '"')


def fields_match_test(fields: Optional[Sequence[str]] = ...):
    """Creates an equivalence test that tests whether the specified fields are equal on both labels.

    Parameters
    ----------
    fields: Optional[Sequence[str]]
        The fields to test or ... if all fields should be tested.

    """

    def fields_match(tested_label: 'data.Label', target_label: 'data.Label') -> bool:
        if fields is not ...:
            return all(
                getattr(tested_label, field) == getattr(target_label, field) for field in fields)
        else:
            return tested_label.shallow_fields_equal(target_label)

    return fields_match


class Accuracy(Metric):
    def __init__(self,
                 name: str = 'accuracy',
                 mode: str = 'equals',
                 print_debug: bool = False,
                 boundary_fuzz: int = 0,
                 fields: Optional[Sequence[str]] = ...,
                 equivalence_test: Optional[Callable[[Any, Any], bool]] = fields_match_test(...)):
        """An accuracy metric with several options for equivalence.

        Parameters
        ----------
        name: str
            An identifier for the metric.
        mode: str
            'equals' - counts as a hit if there is one and only one label at the same location in
            the tested index as the target index and it has the same values for its fields
            'location' - counts as a hit if there is one and only one label at the same location in
            the tested index as the target index.
            'any' - counts as a hit if there is one or more labels at the same location with the
            same values for its fields.
        print_debug: bool
            If true will print debug information about the misses.
        boundary_fuzz: int
            How different the target label boundaries can be from the tested boundaries before it
            doesn't count as a match.
        equivalence_test: callable
            A function which takes two argument labels, and returns true if the labels are
            equivalent for the purpose of the test.
        """
        self.correct = 0
        self.total = 0
        self.name = name
        self.mode = mode
        self.print_debug = print_debug
        self.boundary_fuzz = boundary_fuzz
        if fields is ...:
            self.equivalence_test = equivalence_test
        else:
            self.equivalence_test = fields_match_test(fields)

    @property
    def value(self) -> float:
        return self.correct / self.total if self.total > 0 else 1.

    def find_candidates(self,
                        tested_index: 'data.LabelIndex',
                        target_label: 'data.Label') -> 'data.LabelIndex':
        if self.boundary_fuzz == 0:
            index = tested_index.at(target_label)
        else:
            index = tested_index.inside(
                target_label.start_index - 1,
                target_label.end_index + 1
            ).covering(
                target_label.start_index + 1,
                target_label.end_index - 1
            )
        return index

    def has_match(self, candidates: 'data.LabelIndex', target_label: 'data.Label') -> bool:
        if self.mode == 'equals':
            return len(candidates) == 1 and self.equivalence_test(candidates[0], target_label)
        elif self.mode == 'location':
            return len(candidates) > 0
        elif self.mode == 'any':
            return any(self.equivalence_test(candidate, target_label) for candidate in candidates)

    def update(self, document: 'mtap.Document',
               tested_index: 'data.LabelIndex',
               target_index: 'data.LabelIndex') -> Any:
        correct = 0
        total = 0
        for target_label in target_index:
            total += 1
            candidates = self.find_candidates(tested_index, target_label)
            if self.has_match(candidates, target_label):
                correct += 1
            elif self.print_debug:
                print_overlapping(document, target_label, tested_index)

        self.correct += correct
        self.total += total
        return correct / total if total != 0 else 1.


def _collect_tokens(index, return_insides=True):
    begins = set()
    insides = set()
    for label in index:
        for i, token in enumerate(tokenization.word_tokenize(label.text, label.start_index)):
            if i == 0:
                begins.add(token)
                if not return_insides:
                    break
            else:
                insides.add(token)

    return begins, insides


class ConfusionMatrix(NamedTuple('ConfusionMatrix',
                                 [('true_positives', float),
                                  ('false_positives', float),
                                  ('false_negatives', float)])):
    """A representation of a confusion matrix.

    Attributes
    ----------
    true_positives (float): Count of true positive examples.
    false_positives (float): Count of false positive examples.
    false_negatives (float): Count of false negative examples.
    precision (float): Ratio of true positives to the total number of positive predictions.
    recall (float): Ratio of true positives to the total number of positive ground truths.
    f1 (float): The harmonic mean of precision and recall.

    """

    def __new__(cls, true_positives: float = 0,
                false_positives: float = 0,
                false_negatives: float = 0):
        self = super().__new__(cls, true_positives, false_positives, false_negatives)
        return self

    def __add__(self, other):
        return self.__class__(self.true_positives + other.true_positives,
                              self.false_positives + other.false_positives,
                              self.false_negatives + other.false_negatives)

    @property
    def precision(self):
        predicted_positives = self.true_positives + self.false_positives
        if predicted_positives == 0:
            return 1
        return self.true_positives / predicted_positives

    @property
    def recall(self):
        ground_truth_positives = self.true_positives + self.false_negatives
        if ground_truth_positives == 0:
            return 1
        return self.true_positives / ground_truth_positives

    @property
    def f1(self):
        divisor = 2 * self.true_positives + self.false_positives + self.false_negatives
        if divisor == 0:
            return 1
        return 2 * self.true_positives / divisor


class FirstTokenConfusion(Metric):
    """A metric which treats the first word token in every label as an example of the positive
    class and calculates the precision, recall, and f1 confusion matrix metrics for that
    positive class. Useful for evaluation of segmentation tasks.

    precision = true positives / (true positives + false positives)
    recall = true positives / (true positives + false negatives)
    f1 = 2 * true positives / (2 * true positives + false positives + false negatives)

    Args
    ----
    name (str): An identifying name for the metric.
    tested_filter (~typing.Callable[[Label], bool]): A filter to apply to the tested index.
    target_filter (~typing.Callable[[Label], bool]): A filter to apply to the target index.
    print_debug (str)
        An argument to print failing examples. 'fp' prints only false positive
        errors, 'fn' prints only false negative errors, 'all' prints both false positive and false
        negative errors
    debug_range (int): The range before and after the example to print.
    debug_handle (TextIO): A text io file handle to print the debug information to.

    Attributes
    ----------
    precision (float): Ratio of true positives to the total number of positive predictions.
    recall (float): Ratio of true positives to the total number of positive ground truths.
    f1 (float): The harmonic mean of precision and recall.
    """

    def __init__(self, name: str = 'first_token_confusion',
                 tested_filter: Callable[['data.Label'], bool] = None,
                 target_filter: Callable[['data.Label'], bool] = None,
                 print_debug: str = None,
                 debug_range: int = 30,
                 debug_handle: TextIO = sys.stdout):
        self.name = name
        self._matrix = ConfusionMatrix()
        self.tested_filter = tested_filter
        self.target_filter = target_filter
        self.print_debug = print_debug
        self.debug_range = debug_range
        self.debug_handle = debug_handle

    @property
    def precision(self) -> float:
        return self._matrix.precision

    @property
    def recall(self) -> float:
        return self._matrix.recall

    @property
    def f1(self) -> float:
        return self._matrix.f1

    def update(self,
               document: 'mtap.Document',
               tested_index: 'data.LabelIndex',
               target_index: 'data.LabelIndex') -> Any:
        if self.tested_filter is not None:
            tested_index = tested_index.filter(self.tested_filter)
        if self.target_filter is not None:
            target_index = target_index.filter(self.target_filter)
        tested_tokens, _ = _collect_tokens(tested_index, return_insides=False)
        target_tokens, _ = _collect_tokens(target_index, return_insides=False)
        false_positives = tested_tokens.difference(target_tokens)
        false_negatives = target_tokens.difference(tested_tokens)
        if self.print_debug in ('fp', 'all'):
            self.debug_handle.write('False Positives\n')
            for false_positive in false_positives:
                _print_example(document.text, false_positive, self.debug_range, self.debug_handle)
            self.debug_handle.write('\n')
        if self.print_debug in ('fn', 'all'):
            self.debug_handle.write('False Negatives\n')
            for false_negative in false_negatives:
                _print_example(document.text, false_negative, self.debug_range, self.debug_handle)
            self.debug_handle.write('\n')
        local = ConfusionMatrix(
            true_positives=len(tested_tokens.intersection(target_tokens)),
            false_negatives=len(false_negatives),
            false_positives=len(false_positives)
        )
        self._matrix += local
        return {
            'precision': local.precision,
            'recall': local.recall,
            'f1': local.f1
        }


def _print_example(text: str,
                   token: Tuple[int, int],
                   debug_range: int,
                   debug_handle: TextIO):
    start, end = token
    print_start = max(0, start - debug_range)
    print_end = min(end + debug_range, len(text))
    text = text[print_start:start] + '{' + text[start:end] + '}' + text[end:print_end]
    text = text.replace('\n', ' ') + '\n'
    debug_handle.write(text)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Internal service discovery magic."""

import abc
from uuid import uuid1

from mtap import constants


class Discovery(abc.ABC):
    def __new__(cls, config, *args, **kwargs):
        if config['discovery'] == 'consul':
            cls = ConsulDiscovery
        return super(Discovery, cls).__new__(cls)

    @abc.abstractmethod
    def register_events_service(self, address, port, version):
        raise NotImplementedError

    @abc.abstractmethod
    def discover_events_service(self, version):
        raise NotImplementedError

    @abc.abstractmethod
    def register_processor_service(self, address, port, processor_name, version):
        raise NotImplementedError

    @abc.abstractmethod
    def discover_processor_service(self, processor_name, version):
        raise NotImplementedError


class ConsulDiscovery(Discovery):
    def __init__(self, config):
        import consul
        host = config['consul.host']
        self.c = consul.Consul(host=host,
                               port=config['consul.port'],
                               scheme=config['consul.scheme'])

    def register_events_service(self, address, port, version):
        name = constants.EVENTS_SERVICE_NAME
        self.c.agent.service.register(name,
                                      port=port,
                                      check={
                                          'grpc': "{}:{}/{}".format(address, port, name),
                                          'interval': "10s",
                                          'status': 'passing'
                                      },
                                      tags=[version])

        def deregister():
            self.c.agent.service.deregister(name)

        return deregister

    def discover_events_service(self, version):
        """Delegates discovery to grpc's dns name resolution.

        """
        name = constants.EVENTS_SERVICE_NAME
        _, services = self.c.health.service(name, tag='v1')
        addresses = []
        if len(services) == 0:
            raise ValueError('No addresses found for events service')
        for service in services:
            addresses.append("{}:{}".format(service['Node']['Address'], service['Service']['Port']))
        return "ipv4:" + ','.join(addresses)

    def register_processor_service(self, address, port, processor_id, version):
        uuid = str(uuid1())
        self.c.agent.service.register(processor_id,
                                      service_id=uuid,
                                      port=port,
                                      check={
                                          'grpc': "{}:{}/{}".format(address, port, processor_id),
                                          'interval': "10s",
                                          'status': 'passing'
                                      },
                                      tags=[constants.PROCESSING_SERVICE_TAG])

        def deregister():
            self.c.agent.service.deregister(service_id=uuid)

        return deregister

    def discover_processor_service(self, processor_name, version):
        tag = constants.PROCESSING_SERVICE_TAG
        _, services = self.c.health.service(processor_name, tag=tag)
        addresses = []
        for service in services:
            addresses.append("{}:{}".format(service['Node']['Address'], service['Service']['Port']))
        return "ipv4:" + ','.join(addresses)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from time import sleep

from datetime import timedelta
from typing import Dict, Any

from mtap import Event, EventProcessor


def test_started_stopwatch():
    class TestProcessor(EventProcessor):
        def process(self, event: Event, params: Dict[str, Any]):
            with self.started_stopwatch('foo') as stopwatch:
                sleep(1 / 1000)
            return stopwatch.duration

    pr = TestProcessor()
    assert pr.process(None, None) >= timedelta(milliseconds=1)


def test_not_started_stopwatch():
    class TestProcessor(EventProcessor):
        def process(self, event: Event, params: Dict[str, Any]):
            with self.unstarted_stopwatch('foo') as stopwatch:
                for _ in range(20):
                    stopwatch.start()
                    sleep(1 / 1000)
                    stopwatch.stop()
            return stopwatch.duration

    pr = TestProcessor()
    assert pr.process(None, None) >= timedelta(milliseconds=20)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc_testing
import pytest

import mtap.data._base
from mtap._events_service import EventsServicer
from mtap.api.v1 import events_pb2

PHASERS = """Maybe if we felt any human loss as keenly as we feel one of those close to us, human 
history would be far less bloody. The Enterprise computer system is controlled by three primary 
main processor cores, cross-linked with a redundant melacortz ramistat, fourteen kiloquad interface 
modules. Our neural pathways have become accustomed to your sensory input patterns. Mr. Worf, you 
do remember how to fire phasers?"""


@pytest.fixture(name='events_server')
def fixture_events_server():
    events_service = EventsServicer()
    yield grpc_testing.server_from_dictionary(
        {
            events_pb2.DESCRIPTOR.services_by_name['Events']: events_service
        },
        grpc_testing.strict_real_time()
    )


def test_OpenEvent(events_server: grpc_testing.Server):
    request = events_pb2.OpenEventRequest(event_id='1')
    response, _, _, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        request,
        None
    ).termination()
    assert response.created is True


def test_OpenEvent_without_id(events_server):
    request = events_pb2.OpenEventRequest(event_id=None)
    _, _, status_code, description = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        request,
        None
    ).termination()

    assert status_code == grpc.StatusCode.INVALID_ARGUMENT
    assert description == 'event_id was not set.'


def test_OpenEvent_duplicate(events_server):
    request = events_pb2.OpenEventRequest(event_id='1')
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        request,
        None
    )
    response, _, _, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        request,
        None
    ).termination()
    assert response.created is False


def test_OpenEvent_only_create(events_server):
    request = events_pb2.OpenEventRequest(event_id='1', only_create_new=True)
    response, _, _, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        request,
        None
    ).termination()
    assert response.created is True


def test_OpenEvent_only_create_duplicate(events_server):
    request = events_pb2.OpenEventRequest(event_id='1', only_create_new=True)
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        request,
        None
    )
    _, _, status_code, description = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        request,
        None
    ).termination()
    assert status_code == grpc.StatusCode.ALREADY_EXISTS


def test_CloseEvent_delete(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['CloseEvent'],
        {},
        events_pb2.CloseEventRequest(event_id='1'),
        None
    )
    response, _, _, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    ).termination()
    assert response.created is True


def test_CloseEvent_no_delete(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['CloseEvent'],
        {},
        events_pb2.CloseEventRequest(event_id='1'),
        None
    )
    response, _, _, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    ).termination()
    assert response.created is False


def test_AddMetadataBadEvent(events_server):
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddMetadata'],
        {},
        events_pb2.AddMetadataRequest(event_id='1', key='foo', value='bar'),
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_AddMetadata(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddMetadata'],
        {},
        events_pb2.AddMetadataRequest(event_id='1', key='foo', value='bar'),
        None
    ).termination()
    assert status_code == grpc.StatusCode.OK


def test_AddMetadata_EmptyKey(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddMetadata'],
        {},
        events_pb2.AddMetadataRequest(event_id='1', key='', value='bar'),
        None
    ).termination()
    assert status_code == grpc.StatusCode.INVALID_ARGUMENT


def test_AddMetadata_NoneKey(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddMetadata'],
        {},
        events_pb2.AddMetadataRequest(event_id='1', key=None, value='bar'),
        None
    ).termination()
    assert status_code == grpc.StatusCode.INVALID_ARGUMENT


def test_GetAllMetadata(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddMetadata'],
        {},
        events_pb2.AddMetadataRequest(event_id='1', key='foo', value='bar'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddMetadata'],
        {},
        events_pb2.AddMetadataRequest(event_id='1', key='baz', value='buh'),
        None
    )
    response, _, _, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetAllMetadata'],
        {},
        events_pb2.GetAllMetadataRequest(event_id='1'),
        None
    ).termination()
    d = dict(response.metadata)
    assert d == {'foo': 'bar', 'baz': 'buh'}


def test_GetAllBinaryDataNames(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddBinaryData'],
        {},
        events_pb2.AddBinaryDataRequest(event_id='1',
                                        binary_data_name='a',
                                        binary_data=b'\xBF\xAF'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddBinaryData'],
        {},
        events_pb2.AddBinaryDataRequest(event_id='1',
                                        binary_data_name='b',
                                        binary_data=b'\xAF\xBF'),
        None
    )
    response, _, _, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetAllBinaryDataNames'],
        {},
        events_pb2.GetAllBinaryDataNamesRequest(event_id='1'),
        None
    ).termination()
    assert 'a' in response.binary_data_names
    assert 'b' in response.binary_data_names
    assert len(response.binary_data_names) == 2


def test_AddGetBinaryData(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddBinaryData'],
        {},
        events_pb2.AddBinaryDataRequest(event_id='1',
                                        binary_data_name='a',
                                        binary_data=b'\xBF\xAF'),
        None
    )
    response, _, _, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetBinaryData'],
        {},
        events_pb2.GetBinaryDataRequest(event_id='1', binary_data_name='a'),
        None
    ).termination()
    assert response.binary_data == b'\xBF\xAF'


def test_AddDocument_bad_event(events_server):
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_AddDocument(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    ).termination()
    assert status_code == grpc.StatusCode.OK


def test_AddDocument_empty_name(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='',
                                      text=PHASERS),
        None
    ).termination()
    assert status_code == grpc.StatusCode.INVALID_ARGUMENT


def test_AddDocument_empty_text(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=''),
        None
    ).termination()
    assert status_code == grpc.StatusCode.OK


def test_AddDocument_exists(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    )
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    ).termination()
    assert status_code == grpc.StatusCode.ALREADY_EXISTS


def test_GetAllDocuments(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=''),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='other',
                                      text=''),
        None
    )
    response, _, _, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetAllDocumentNames'],
        {},
        events_pb2.GetAllDocumentNamesRequest(event_id='1'),
        None
    ).termination()
    assert 'plaintext' in response.document_names
    assert 'other' in response.document_names
    assert len(response.document_names) == 2


def test_GetDocumentText_bad_event(events_server):
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetDocumentText'],
        {},
        events_pb2.GetDocumentTextRequest(event_id='1',
                                          document_name='plaintext'),
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_GetDocumentText_bad_document(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetDocumentText'],
        {},
        events_pb2.GetDocumentTextRequest(event_id='1',
                                          document_name='plaintext'),
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_GetDocumentText(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    )
    response, _, _, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetDocumentText'],
        {},
        events_pb2.GetDocumentTextRequest(event_id='1',
                                          document_name='plaintext'),
        None
    ).termination()
    assert response.text == PHASERS


def test_GetLabelIndicesInfo_bad_event(events_server):
    request = events_pb2.GetLabelIndicesInfoRequest(event_id='1', document_name='plaintext')

    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetLabelIndicesInfo'],
        {},
        request,
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_GetLabelIndicesInfo_bad_document(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    request = events_pb2.GetLabelIndicesInfoRequest(event_id='1', document_name='plaintext')

    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetLabelIndicesInfo'],
        {},
        request,
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_GetLabelIndicesInfo(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    )
    request = events_pb2.AddLabelsRequest(event_id='1',
                                          document_name='plaintext',
                                          index_name='labels')
    s = request.generic_labels.labels.add()
    s.start_index = 15
    s.end_index = 20
    s.fields['some_other_field'] = 'blah'
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddLabels'],
        {},
        request,
        None
    ).termination()
    assert status_code == grpc.StatusCode.OK
    request = events_pb2.GetLabelIndicesInfoRequest(event_id='1', document_name='plaintext')
    res, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetLabelIndicesInfo'],
        {},
        request,
        None
    ).termination()
    assert status_code == grpc.StatusCode.OK
    assert len(res.label_index_infos) == 1
    assert res.label_index_infos[0].index_name == 'labels'
    assert res.label_index_infos[0].type == events_pb2.GetLabelIndicesInfoResponse.LabelIndexInfo.GENERIC


def test_AddLabels_bad_event(events_server):
    request = events_pb2.AddLabelsRequest(event_id='1',
                                          document_name='plaintext',
                                          index_name='labels')
    label = request.generic_labels.labels.add()
    label.start_index = 15
    label.end_index = 20
    label.fields['some_other_field'] = 'blah'

    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddLabels'],
        {},
        request,
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_AddLabels_bad_document(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    request = events_pb2.AddLabelsRequest(event_id='1',
                                          document_name='plaintext',
                                          index_name='labels')
    label = request.generic_labels.labels.add()
    label.start_index = 15
    label.end_index = 20
    label.fields['some_other_field'] = 'blah'

    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddLabels'],
        {},
        request,
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_AddLabels_bad_index_name(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    )
    request = events_pb2.AddLabelsRequest(event_id='1',
                                          document_name='plaintext',
                                          index_name='')
    label = request.generic_labels.labels.add()
    label.start_index = 15
    label.end_index = 20
    label.fields['some_other_field'] = 'blah'

    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddLabels'],
        {},
        request,
        None
    ).termination()
    assert status_code == grpc.StatusCode.INVALID_ARGUMENT


def test_AddLabels(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    )
    request = events_pb2.AddLabelsRequest(event_id='1',
                                          document_name='plaintext',
                                          index_name='labels')
    s = request.generic_labels.labels.add()
    s.start_index = 15
    s.end_index = 20
    s.fields['some_other_field'] = 'blah'
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddLabels'],
        {},
        request,
        None
    ).termination()
    assert status_code == grpc.StatusCode.OK


def test_AddLabels_no_labels(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    )
    request = events_pb2.AddLabelsRequest(event_id='1',
                                          document_name='plaintext',
                                          index_name='labels')
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddLabels'],
        {},
        request,
        None
    ).termination()
    assert status_code == grpc.StatusCode.OK


def test_GetLabels_bad_event(events_server):
    r = events_pb2.GetLabelsRequest(event_id='1', document_name='plaintext', index_name='labels')
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetLabels'],
        {},
        r,
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_GetLabels_bad_document_name(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    r = events_pb2.GetLabelsRequest(event_id='1', document_name='plaintext', index_name='labels')
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetLabels'],
        {},
        r,
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_GetLabels_bad_index_name(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    )
    r = events_pb2.GetLabelsRequest(event_id='1', document_name='plaintext', index_name='labels')
    _, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetLabels'],
        {},
        r,
        None
    ).termination()
    assert status_code == grpc.StatusCode.NOT_FOUND


def test_GetLabels_no_labels(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddLabels'],
        {},
        events_pb2.AddLabelsRequest(event_id='1',
                                    document_name='plaintext',
                                    index_name='labels'),
        None
    )
    req = events_pb2.GetLabelsRequest(event_id='1', document_name='plaintext', index_name='labels')
    res, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetLabels'],
        {},
        req,
        None
    ).termination()
    assert len(res.generic_labels.labels) == 0


def test_GetLabels(events_server):
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['OpenEvent'],
        {},
        events_pb2.OpenEventRequest(event_id='1'),
        None
    )
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddDocument'],
        {},
        events_pb2.AddDocumentRequest(event_id='1',
                                      document_name='plaintext',
                                      text=PHASERS),
        None
    )
    request = events_pb2.AddLabelsRequest(event_id='1',
                                          document_name='plaintext',
                                          index_name='labels')
    s = request.generic_labels.labels.add()
    s.start_index = 15
    s.end_index = 20
    s.fields['some_other_field'] = 'blah'
    events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['AddLabels'],
        {},
        request,
        None
    )

    req = events_pb2.GetLabelsRequest(event_id='1', document_name='plaintext', index_name='labels')
    res, _, status_code, _ = events_server.invoke_unary_unary(
        events_pb2.DESCRIPTOR.services_by_name['Events'].methods_by_name['GetLabels'],
        {},
        req,
        None
    ).termination()
    assert len(res.generic_labels.labels) == 1
    label = res.generic_labels.labels[0]
    assert label.start_index == 15
    assert label.end_index == 20
    assert label.fields['some_other_field'] == 'blah'



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import os
import pathlib
import signal
import subprocess
import typing

import grpc


__all__ = [
    'find_free_port',
    'subprocess_events_server',
    'read_address',
    'write_address_file',
    'tokenization'
]


def find_free_port() -> int:
    """Borrowed from ``https://stackoverflow.com/questions/1365265``.

    Returns
    -------
    int
        A free port which can be bound to.

    """
    import socket
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


@contextlib.contextmanager
def subprocess_events_server(port: typing.Optional[int] = None,
                             cwd: typing.Optional[pathlib.Path] = None,
                             config_path: typing.Optional[typing.Union[pathlib.Path, str]] = None,
                             register: bool = False) -> str:
    """Context manager which launches a events server on a subprocess and yields the address.

    Parameters
    ----------
    port: optional int
        A port to bind the events server to.
    cwd: optional Path
        A current working directory to use.
    config_path: optional Path
        A path to a configuration file to use.
    register: bool
        Whether to register the events server with service discovery.

    Returns
    -------
    str
        The address that the events server us running on.

    """
    if cwd is None:
        cwd = pathlib.Path.cwd()
    env = dict(os.environ)
    if config_path is not None:
        env['MTAP_CONFIG'] = str(config_path)
    if port is None:
        port = find_free_port()
    address = '127.0.0.1:' + str(port)
    cmd = ['python', '-m', 'mtap', 'events', '-p', str(port)]
    if register:
        cmd += ['--register']
    p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT, cwd=str(cwd), env=env)
    try:
        with grpc.insecure_channel(address, [('grpc.enable_http_proxy', False)]) as channel:
            future = grpc.channel_ready_future(channel)
            future.result(timeout=10)
        yield address
    finally:
        p.send_signal(signal.SIGINT)
        try:
            stdout, _ = p.communicate(timeout=5)
        except subprocess.TimeoutExpired:
            p.kill()
            stdout, _ = p.communicate()
        print(stdout.decode('utf-8'))
        print("python events exited with code: ", p.returncode)


def write_address_file(address: str, pid: typing.Optional[str] = None) -> pathlib.Path:
    """Writes the address file (a file containing just the address for a processor). This is a file
    used to communicate a service's port (potentially chosen randomly) back to the script that
    launched the service.

    Args:
        address (str): The host.
        pid (str, optional): The process ID. If omitted will use `os.getpid()` to retrieve the pid.

    Returns:
        str: The path to the file.

    """
    directory = mtap_home() / 'addresses'
    directory.mkdir(parents=True, exist_ok=True)
    if pid is None:
        pid = os.getpid()
    address_path = directory / '{}.address'.format(pid)
    with address_path.open('w') as fio:
        fio.write(address)
    return address_path


def read_address(pid: str) -> str:
    """Reads the address for a service with a specified PID.

    Args:
        pid (str): The service's PID.

    Returns:
        str: The address.

    """
    directory = mtap_home() / 'addresses'
    address_path = directory / '{}.address'.format(pid)
    with address_path.open('r') as fin:
        txt = fin.read()
    return txt


def mtap_home() -> pathlib.Path:
    """Provides a path to the MTAP home directory.

    Returns:
        pathlib.Path: The home directory.

    """
    import os
    try:
        home_env = os.environ['MTAP_HOME']
        return pathlib.Path(home_env)
    except KeyError:
        return pathlib.Path.home() / '.mtap'



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An example document processor."""

from typing import Dict, Any, Optional

import mtap
from mtap.processing.descriptions import parameter, labels, label_property


@mtap.processor('mtap-example-processor-python',
                human_name="Python Example Processor",
                description="counts the number of times the letters a and b occur in a document",
                parameters=[
                    parameter('do_work', required=True, data_type='bool',
                              description="Whether the processor should do anything.")
                ],
                outputs=[
                    labels('mtap.examples.letter_counts',
                           properties=[label_property('letter', data_type='str'),
                                       label_property('count', data_type='int')])
                ])
class ExampleProcessor(mtap.DocumentProcessor):
    """Does some labeling of the counts of the letter 'a' and 'b' in a document.
    """

    def process_document(self,
                         document: mtap.Document,
                         params: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        if params['do_work']:
            with self.started_stopwatch('fetch_time'):
                text = document.text

            a_count = text.count('a')
            b_count = text.count('b')

            with document.get_labeler('mtap.examples.letter_counts') as label_letter_count:
                label_letter_count(start_index=0, end_index=len(document.text), letter='a',
                                   count=a_count)
                label_letter_count(start_index=0, end_index=len(document.text), letter='b',
                                   count=b_count)

        return {'answer': 42}


if __name__ == '__main__':
    mtap.run_processor(ExampleProcessor(), mp=True)



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from tempfile import TemporaryFile

import mtap
from mtap import Event, Document
from mtap.io.serialization import JsonSerializer


def test_json_serializer():
    event = Event(event_id='1')
    event.metadata['foo'] = "bar"
    document = Document('plaintext', text='Some text.')
    event.add_document(document)
    one = mtap.GenericLabel(start_index=0, end_index=5, x=10)
    two = mtap.GenericLabel(start_index=6, end_index=10, x=15)
    document.add_labels('one', [one,
                                two])
    document.add_labels('two', [mtap.GenericLabel(start_index=0, end_index=25, a='b', b=one),
                                mtap.GenericLabel(start_index=26, end_index=42, a='c', b=two)])
    document.add_labels('three', [
        mtap.GenericLabel(start_index=0, end_index=10, foo=True),
        mtap.GenericLabel(start_index=11, end_index=15, foo=False)
    ], distinct=True)

    with TemporaryFile('w+') as tf:
        JsonSerializer.event_to_file(event, tf)
        tf.flush()
        tf.seek(0)
        e = JsonSerializer.file_to_event(tf)

    assert e.event_id == event.event_id
    assert e.metadata['foo'] == 'bar'
    d = e.documents['plaintext']
    assert d.text == document.text
    index_one = d.labels['one']
    assert index_one == [one, two]
    index_two = d.labels['two']
    assert index_two == [mtap.GenericLabel(start_index=0, end_index=25, a='b', b=one),
                         mtap.GenericLabel(start_index=26, end_index=42, a='c', b=two)]
    index_three = d.labels['three']
    assert index_three == [mtap.GenericLabel(start_index=0, end_index=10, foo=True),
                           mtap.GenericLabel(start_index=11, end_index=15, foo=False)]



# Copyright 2019 Regents of the University of Minnesota.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Any

import grpc
import grpc_testing
import pytest

from mtap import processor, Document
from mtap.api.v1 import processing_pb2
from mtap.processing import DocumentProcessor, _runners
from mtap.processing.descriptions import parameter, labels, label_property
from mtap.processing._service import _ProcessorServicer


@processor('mtap-test-processor',
           description='Processor desc.',
           parameters=[
               parameter('a_param', required=True, data_type='bool',
                         description="desc.")
           ],
           inputs=[
               labels('input_index', properties=[label_property('bar', data_type='bool')])
           ],
           outputs=[
               labels('output_index',
                      description='desc.',
                      properties=[label_property('foo', data_type='str', nullable=True,
                                                      description='A label property.')])
           ])
class ExampleTestProcessor(DocumentProcessor):
    def process_document(self, document: Document, params: Dict[str, Any]):
        pass


@pytest.fixture(name='processor_servicer')
def fixture_processor_servicer():
    runner = _runners.ProcessorRunner(ExampleTestProcessor(), client=None)
    processor_service = _ProcessorServicer(config={}, runner=runner, address='',
                                           health_servicer=None)
    yield grpc_testing.server_from_dictionary(
        {
            processing_pb2.DESCRIPTOR.services_by_name['Processor']: processor_service
        },
        grpc_testing.strict_real_time()
    )


def test_GetInfo(processor_servicer):
    request = processing_pb2.GetInfoRequest(processor_id='mtap-example-processor-python')
    resp, _, status_code, _ = processor_servicer.invoke_unary_unary(
        processing_pb2.DESCRIPTOR.services_by_name['Processor'].methods_by_name['GetInfo'],
        {},
        request,
        None
    ).termination()

    assert status_code == grpc.StatusCode.OK
    assert resp.metadata['name'] == 'mtap-test-processor'
    assert len(resp.metadata['parameters']) == 1
    assert resp.metadata['parameters'][0]['name'] == 'a_param'
    assert resp.metadata['parameters'][0]['required']
    assert resp.metadata['parameters'][0]['data_type'] == 'bool'
    assert resp.metadata['parameters'][0]['description'] == 'desc.'
    assert len(resp.metadata['inputs']) == 1
    assert resp.metadata['inputs'][0]['name'] == 'input_index'
    assert len(resp.metadata['inputs'][0]['properties']) == 1
    assert resp.metadata['inputs'][0]['properties'][0]['name'] == 'bar'
    assert resp.metadata['inputs'][0]['properties'][0]['data_type'] == 'bool'
    assert len(resp.metadata['outputs']) == 1
    assert resp.metadata['outputs'][0]['name'] == 'output_index'
    assert resp.metadata['outputs'][0]['description'] == 'desc.'
    assert len(resp.metadata['outputs'][0]['properties']) == 1
    assert resp.metadata['outputs'][0]['properties'][0]['name'] == 'foo'
    assert resp.metadata['outputs'][0]['properties'][0]['data_type'] == 'str'
    assert resp.metadata['outputs'][0]['properties'][0]['nullable']
    assert resp.metadata['outputs'][0]['properties'][0]['description'] == 'A label property.'
